import numpy as np

from .steerable_basis import IrrepBasis
from .steerable_filters_basis import SteerableFiltersBasis

from escnn.group import *

import torch

from typing import Union, Tuple, Dict, Iterable, List
from collections import defaultdict
from itertools import chain

__all__ = [
    "WignerEckartBasis",
    "RestrictedWignerEckartBasis"
]


class WignerEckartBasis(IrrepBasis):
    
    def __init__(self,
                 basis: SteerableFiltersBasis,
                 in_irrep: Union[str, IrreducibleRepresentation, int],
                 out_irrep: Union[str, IrreducibleRepresentation, int],
        ):
        r"""
        
        Solves the kernel constraint for a pair of input and output :math:`G`-irreps by using the Wigner-Eckart theorem
        described in Theorem 2.1 of
        `A Program to Build E(N)-Equivariant Steerable CNNs <https://openreview.net/forum?id=WE4qe9xlnQw>`_
        (see also
        `A Wigner-Eckart Theorem for Group Equivariant Convolution Kernels <https://arxiv.org/abs/2010.10952>`_
        ).
        The method relies on a :math:`G`-Steerable basis of scalar functions over the base space.

        Args:
            basis (SteerableFiltersBasis): a `G`-steerable basis for scalar functions over the base space
            in_repr (IrreducibleRepresentation): the input irrep
            out_repr (IrreducibleRepresentation): the output irrep

        """
        
        group = basis.group

        in_irrep = group.irrep(*group.get_irrep_id(in_irrep))
        out_irrep = group.irrep(*group.get_irrep_id(out_irrep))
        
        assert in_irrep.group == group
        assert out_irrep.group == group
        assert in_irrep.group == out_irrep.group

        self.m = in_irrep.id
        self.n = out_irrep.id
        
        _js = group._tensor_product_irreps(self.m, self.n)

        _js = [
            (j, jJl)
            for j, jJl in _js
            if basis.multiplicity(j) > 0
        ]

        dim = 0
        self._dim_harmonics = {}
        self._jJl = {}
        for j, jJl in _js:
            self._dim_harmonics[j] = basis.multiplicity(j) * jJl * group.irrep(*j).sum_of_squares_constituents
            self._jJl[j] = jJl
            dim += self._dim_harmonics[j]

        super(WignerEckartBasis, self).__init__(basis, in_irrep, out_irrep, dim, harmonics=[_j for _j, _ in _js])

        # SteerableFiltersBasis: a `G`-steerable basis for scalar functions over the base space
        self.basis = basis

        _coeff = [
            torch.einsum(
                # 'mnsi,koi->mnkso',
                'mnsi,koi->ksmno',
                torch.tensor(group._clebsh_gordan_coeff(self.n, self.m, j), dtype=torch.float32),
                torch.tensor(group.irrep(*j).endomorphism_basis(), dtype=torch.float32),
            ) for j in self.js
        ]
        
        for b, j in enumerate(self.js):
            coeff = _coeff[b]
            assert self._jJl[j] == coeff.shape[1]
            self.register_buffer(f'coeff_{b}', coeff)

    def coeff(self, idx: int) -> torch.Tensor:
        return getattr(self, f'coeff_{idx}')

    def sample_harmonics(self, points: Dict[Tuple, torch.Tensor], out: Dict[Tuple, torch.Tensor] = None) -> Dict[Tuple, torch.Tensor]:
        if out is None:
            out = {
                j: torch.zeros(
                    (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1]),
                    device=points[j].device, dtype=points[j].dtype
                )
                for j in self.js
            }
    
        for j in self.js:
            if j in out:
                assert out[j].shape == (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1]), (
                    out[j].shape, points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1]
                )

        for b, j in enumerate(self.js):
            if j not in out:
                continue
            coeff = self.coeff(b)
            
            jJl = coeff.shape[1]
        
            Ys = points[j]

            out[j].view((
                Ys.shape[0],
                self.group.irrep(*j).sum_of_squares_constituents, jJl,
                Ys.shape[1],
                self.out_irrep.size, self.in_irrep.size,
            ))[:] = torch.einsum(
                # 'Nnksm,miS->NnksiS',
                'kspnm,qim->qksipn',
                coeff, Ys,
            )
        
        return out

    def dim_harmonic(self, j: Tuple) -> int:
        if j in self._dim_harmonics:
            return self._dim_harmonics[j]
        else:
            return 0

    def attrs_j_iter(self, j: Tuple) -> Iterable:
        
        if self.dim_harmonic(j) == 0:
            return
        
        idx = self._start_index[j]
        
        j_attr = {
            'irrep:'+k: v
            for k, v in self.group.irrep(*j).attributes.items()
        }
        steerable_basis_j_attr = list(self.basis.steerable_attrs_j_iter(j))
        
        for k in range(self.group.irrep(*j).sum_of_squares_constituents):
            for s in range(self._jJl[j]):
                for i, attr_i in enumerate(steerable_basis_j_attr):
                    attr = j_attr.copy()
                    attr.update(**attr_i)
                    attr["idx"] = idx
                    attr["j"] = j
                    attr["i"] = i
                    attr["s"] = s
                    attr["k"] = k
                    idx += 1
                
                    yield attr

    def attrs_j(self, j: Tuple, idx) -> Dict:
        assert 0 <= idx < self.dim_harmonic(j)
        
        full_idx = self._start_index[j] + idx

        dim = self.basis.multiplicity(j)

        attr = {
            'irrep:'+k: v
            for k, v in self.group.irrep(*j).attributes.items()
        }

        i = idx % dim
        attr_i = self.basis.steerable_attrs_j(j, i)

        attr.update(**attr_i)
        attr["idx"] = full_idx
        attr["j"] = j
        attr["i"] = i
        attr["s"] = (idx // dim) % self._jJl[j]
        attr["k"] = idx // (dim * self._jJl[j])
        
        return attr

    def __getitem__(self, idx):
        assert 0 <= idx < self.dim
    
        i = idx
        for j in self.js:
            dim = self.dim_harmonic(j)
            if i < dim:
                break
            else:
                i -= dim
        
        return self.attrs_j(j, i)
        
    def __iter__(self):
        for j in self.js:
            for attr in self.attrs_j_iter(j):
                yield attr
        # return chain(self.attrs_j_iter(j) for j in self.js)

    def __eq__(self, other):
        if not isinstance(other, WignerEckartBasis):
            return False
        elif self.basis != other.basis or self.in_irrep != other.in_irrep or self.out_irrep != other.out_irrep:
            # TODO check isomorphism too!
            return False
        elif len(self.js) != len(other.js):
            return False
        else:
            for b, (j, i) in enumerate(zip(self.js, other.js)):
                if j!=i or not torch.allclose(self.coeff(b), other.coeff(b)):
                    return False
            return True

    def __hash__(self):
        return hash(self.basis) + hash(self.in_irrep) + hash(self.out_irrep) + hash(tuple(self.js))

    _cached_instances = {}
    
    @classmethod
    def _generator(cls,
                   basis: SteerableFiltersBasis,
                   psi_in: Union[IrreducibleRepresentation, Tuple],
                   psi_out: Union[IrreducibleRepresentation, Tuple],
                   **kwargs) -> 'IrrepBasis':
    
        assert len(kwargs) == 0
        
        psi_in = basis.group.irrep(*basis.group.get_irrep_id(psi_in))
        psi_out = basis.group.irrep(*basis.group.get_irrep_id(psi_out))
        
        key = (basis, psi_in.id, psi_out.id)
        
        if key not in cls._cached_instances:
            cls._cached_instances[key] = WignerEckartBasis(basis, in_irrep=psi_in, out_irrep=psi_out)
        return cls._cached_instances[key]


class RestrictedWignerEckartBasis(IrrepBasis):
    
    def __init__(self,
                 basis: SteerableFiltersBasis,
                 sg_id: Tuple,
                 in_irrep: Union[str, IrreducibleRepresentation, int],
                 out_irrep: Union[str, IrreducibleRepresentation, int],
                 ):
        r"""

        Solves the kernel constraint for a pair of input and output :math:`G`-irreps by using the Wigner-Eckart theorem
        described in Theorem 2.1 of
        `A Program to Build E(N)-Equivariant Steerable CNNs  <https://openreview.net/forum?id=WE4qe9xlnQw>`_.
        This method implicitly constructs the required :math:`G`-steerable basis for scalar functions on the base space
        from a :math:`G'`-steerable basis, with :math:`G' > G` a larger group, according to Equation 5 from the same
        paper.

        The equivariance group :math:`G < G'` is identified by the input id ``sg_id``.

        .. warning::
            Note that the group :math:`G'` associated with ``basis`` is generally not the same as the group :math:`G`
            associated with ``in_irrep`` and ``out_irrep`` and which the resulting kernel basis is equivariant to.

        Args:
            basis (SteerableFiltersBasis): :math:`G'`-steerable basis for scalar filters
            sg_id (tuple): id of :math:`G` as a subgroup of :math:`G'`.
            in_repr (IrreducibleRepresentation): the input `G`-irrep
            out_repr (IrreducibleRepresentation): the output `G`-irrep

        """

        # the larger group G'
        _G = basis.group

        G = _G.subgroup(sg_id)[0]
        # Group: the smaller equivariance group G
        self.group = G
        self.sg_id = sg_id

        in_irrep = G.irrep(*G.get_irrep_id(in_irrep))
        out_irrep = G.irrep(*G.get_irrep_id(out_irrep))

        assert in_irrep.group == G
        assert out_irrep.group == G
        assert in_irrep.group == out_irrep.group
        
        self.m = in_irrep.id
        self.n = out_irrep.id

        # irreps of G in the decomposition of the tensor product of in_irrep and out_irrep
        _js_G = [
            j for j, _ in
            G._tensor_product_irreps(self.m, self.n)
        ]
        
        _js = set()
        _js_restriction = defaultdict(list)
        
        # for each harmonic j' to consider
        for _j in set(_j for _j, _ in basis.js):
            if basis.multiplicity(_j) == 0:
                continue
                
            # restrict the corresponding G' irrep j' to G
            _j_G = _G.irrep(*_j).restrict(sg_id)
            
            # for each G-irrep j in the tensor product decomposition of in_irrep and out_irrep
            for j in _js_G:
                # irrep-decomposition coefficients of j in j'
                id_coeff = []
                p = 0
                # for each G-irrep i in the restriction of j' to G
                for i in _j_G.irreps:
                    size = G.irrep(*i).size
                    # if the restricted irrep contains one of the irreps in the tensor product
                    if i == j:
                        id_coeff.append(
                            _j_G.change_of_basis_inv[p:p+size, :]
                        )
                    
                    p += size
                
                # if the G irrep j appears in the restriction of the G'-irrep j',
                # store its irrep-decomposition coefficients
                if len(id_coeff) > 0:
                    id_coeff = np.stack(id_coeff, axis=-1)
                    _js.add(_j)
                    _js_restriction[_j].append((j, id_coeff))

        _js = sorted(list(_js))

        self._js_restriction = {}
        self._dim_harmonics = {}
        _coeffs = {}
        dim = 0
        for _j in _js:
            Y_size = _G.irrep(*_j).size
            coeff = [
                torch.einsum(
                    # 'nmsi,kji,jyt->nmksty',
                    'nmsi,kji,jyt->kstnmy',
                    torch.tensor(G._clebsh_gordan_coeff(self.n, self.m, j), dtype=torch.float32),
                    torch.tensor(G.irrep(*j).endomorphism_basis(), dtype=torch.float32),
                    torch.tensor(id_coeff, dtype=torch.float32),
                ).reshape((-1, out_irrep.size, in_irrep.size, Y_size))
                for j, id_coeff in _js_restriction[_j]
            ]
            
            _coeffs[_j] = torch.cat(coeff, dim=0)
            self._js_restriction[_j] = [(j, id_coeff.shape[2]) for j, id_coeff in _js_restriction[_j]]
            self._dim_harmonics[_j] = _coeffs[_j].shape[0]
            dim += self._dim_harmonics[_j] * basis.multiplicity(_j)

        super(RestrictedWignerEckartBasis, self).__init__(basis, in_irrep, out_irrep, dim, harmonics=_js)

        # SteerableFiltersBasis: a `G'`-steerable basis for scalar functions over the base space, for the larger
        # group `G' > G`
        self.basis = basis

        for b, _j in enumerate(self.js):
            self.register_buffer(f'coeff_{b}', _coeffs[_j])

    def coeff(self, idx: int) -> torch.Tensor:
        return getattr(self, f'coeff_{idx}')

    def sample_harmonics(self, points: Dict[Tuple, torch.Tensor], out: Dict[Tuple, torch.Tensor] = None) -> Dict[
        Tuple, torch.Tensor]:
        if out is None:
            out = {
                j: torch.zeros(
                    (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1]),
                    device=points[j].device, dtype=points[j].dtype
                )
                for j in self.js
            }
    
        for j in self.js:
            if j in out:
                assert out[j].shape == (points[j].shape[0], self.dim_harmonic(j), self.shape[0], self.shape[1])
    
        for b, j in enumerate(self.js):
            if j not in out:
                continue
            coeff = self.coeff(b)
        
            Ys = points[j]
            out[j].view((
                Ys.shape[0], coeff.shape[0], Ys.shape[1],
                self.out_irrep.size, self.in_irrep.size,
            ))[:] = torch.einsum(
                'dpnm,sim->sdipn',
                coeff, Ys,
            )
    
        return out

    def dim_harmonic(self, j: Tuple) -> int:
        if j in self._dim_harmonics:
            return self.basis.multiplicity(j) * self._dim_harmonics[j]
        else:
            return 0

    def attrs_j_iter(self, j: Tuple) -> Iterable:
        
        if self.dim_harmonic(j) == 0:
            return
        
        idx = self._start_index[j]

        steerable_basis_j_attr = list(self.basis.steerable_attrs_j_iter(j))

        j_attr = {
            'irrep:'+k: v
            for k, v in self.basis.group.irrep(*j).attributes.items()
        }

        count = 0
        for _j, _jj in self._js_restriction[j]:
            _jJl = self.group._clebsh_gordan_coeff(self.n, self.m, _j).shape[2]
            K = self.group.irrep(*_j).sum_of_squares_constituents
            
            for k in range(K):
                for s in range(_jJl):
                    for t in range(_jj):
                        for i, attr_i in enumerate(steerable_basis_j_attr):
                            attr = j_attr.copy()
                            attr.update(**attr_i)
                            attr["idx"] = idx
                            attr["j"] = j
                            attr["_j"] = _j
                            attr["i"] = i
                            attr["t"] = t
                            attr["s"] = s
                            attr["k"] = k

                            assert idx < self.dim
                            assert count < self.dim_harmonic(j), (count, self.dim_harmonic(j))

                            idx += 1
                            count += 1
                        
                            yield attr

        assert count == self.dim_harmonic(j), (count, self.dim_harmonic(j))

    def attrs_j(self, j: Tuple, idx) -> Dict:
        assert 0 <= idx < self.dim_harmonic(j)
        
        full_idx = self._start_index[j] + idx

        dim = self.basis.multiplicity(j)

        for _j, _jj in self._js_restriction[j]:
            _jJl = self.group._clebsh_gordan_coeff(self.n, self.m, _j).shape[2]
            K = self.group.irrep(*_j).sum_of_squares_constituents
            
            d = _jj * _jJl * K * dim
            
            if idx >= d:
                idx -= d
            else:
                break

        i = idx % dim
        attr_i = self.basis.steerable_attrs_j(j, i)

        attr = {
            'irrep:'+k: v
            for k, v in self.basis.group.irrep(*j).attributes.items()
        }
        attr.update(**attr_i)
        attr["idx"] = full_idx
        attr["j"] = j
        attr["_j"] = _j
        attr["i"] = i
        attr["t"] = (idx // dim) % _jj
        attr["s"] = (idx // (dim * _jj)) % _jJl
        attr["k"] = idx // (dim * _jj * _jJl)
        return attr

    def __getitem__(self, idx):
        assert 0 <= idx < self.dim
    
        i = idx
        for j in self.js:
            dim = self.dim_harmonic(j)
            if i < dim:
                break
            else:
                i -= dim
    
        return self.attrs_j(j, i)

    def __iter__(self) -> Iterable:
        for j in self.js:
            for attr in self.attrs_j_iter(j):
                yield attr
        # return chain(self.attrs_j_iter(j) for j in self.js)

    def __eq__(self, other):
        if not isinstance(other, RestrictedWignerEckartBasis):
            return False
        elif self.basis != other.basis or self.sg_id != other.sg_id or self.in_irrep != other.in_irrep or self.out_irrep != other.out_irrep:
            # TODO check isomorphism too!
            return False
        elif len(self.js) != len(other.js):
            return False
        else:
            for b, (j, i) in enumerate(zip(self.js, other.js)):
                if j!=i or not torch.allclose(self.coeff(b), other.coeff(b)):
                    return False
            return True

    def __hash__(self):
        return hash(self.basis) + hash(self.sg_id) + hash(self.in_irrep) + hash(self.out_irrep) + hash(tuple(self.js))

    _cached_instances = {}
    
    @classmethod
    def _generator(cls,
                   basis: SteerableFiltersBasis,
                   psi_in: Union[IrreducibleRepresentation, Tuple],
                   psi_out: Union[IrreducibleRepresentation, Tuple],
                   **kwargs) -> 'IrrepBasis':
    
        assert len(kwargs) == 1
        assert 'sg_id' in kwargs

        G, _, _ = basis.group.subgroup(kwargs['sg_id'])
        psi_in = G.irrep(*G.get_irrep_id(psi_in))
        psi_out = G.irrep(*G.get_irrep_id(psi_out))

        key = (
            basis, psi_in.id, psi_out.id,
            kwargs['sg_id']
        )

        if key not in cls._cached_instances:
            cls._cached_instances[key] = RestrictedWignerEckartBasis(
                basis,
                sg_id=kwargs['sg_id'],
                in_irrep=psi_in,
                out_irrep=psi_out,
            )
        
        return cls._cached_instances[key]





from escnn.kernels import KernelBasis, EmptyBasisException
from .basismanager import BasisManager

from typing import Callable, Dict, List, Iterable, Union

import torch
import numpy as np

__all__ = ["SingleBlockBasisExpansion", "block_basisexpansion"]


class SingleBlockBasisExpansion(torch.nn.Module, BasisManager):
    
    def __init__(self,
                 basis: KernelBasis,
                 points: np.ndarray,
                 mask: np.ndarray = None
                 ):
        r"""
        
        Basis expansion method for a single contiguous block, i.e. for kernels whose input type and output type contain
        only fields of one type.
        
        Args:
            basis (KernelBasis): analytical basis to sample
            points (np.ndarray): points where the analytical basis should be sampled
            mask (np.ndarray, optional): binary mask to select only a subset of the basis elements.
                                         By default (``None``), all elements are kept.
            
        """

        super(SingleBlockBasisExpansion, self).__init__()

        # This is a hack to prevent PyTorch to register basis as a submodule
        # This is needed since a KernelBasis is a torch.nn.Module, but we only need it in the __init__ to construct the
        # filter basis. We keep this basis mostly for debugging or inspection purpose (e.g. to generate the basis
        # attributes), but we don't want it to appear among the sub-modules of this class (e.g. to prevent it is loaded
        # on CUDA, wasting memory, when this module is moved).
        object.__setattr__(self, 'basis', basis)

        if mask is None:
            mask = np.ones(len(basis), dtype=bool)
            
        assert mask.shape == (len(basis),) and mask.dtype == bool
        
        if not mask.any():
            raise EmptyBasisException

        # we need to know the real output size of the basis elements (i.e. without the change of basis and the padding)
        # to perform the normalization
        sizes = []
        for attr in basis:
            sizes.append(attr["shape"][0])

        # sample the basis on the grid
        # basis has shape (p, k, o, i)
        # permute to (k, o, i, p)
        sampled_basis = basis.sample(torch.tensor(
            points, dtype=torch.float32
        )).permute(1, 2, 3, 0)

        # normalize the basis
        sizes = torch.tensor(sizes, dtype=sampled_basis.dtype)
        sampled_basis = normalize_basis(sampled_basis, sizes)

        # discard the basis which are close to zero everywhere
        norms = (sampled_basis ** 2).reshape(sampled_basis.shape[0], -1).sum(1) > 1e-2
        mask = torch.tensor(mask) & norms
        if not mask.any():
            raise EmptyBasisException
        sampled_basis = sampled_basis[mask, ...]
        self._mask = mask
        
        # register the bases tensors as parameters of this module
        self.register_buffer('sampled_basis', sampled_basis)
            
    def forward(self, weights: torch.Tensor) -> torch.Tensor:
    
        assert len(weights.shape) == 2 and weights.shape[1] == self.dimension()
    
        # expand the current subset of basis vectors and set the result in the appropriate place in the filter
        return torch.einsum('boi...,kb->koi...', self.sampled_basis, weights)

    def get_element_info(self, id: int) -> Dict:
        idx = 0
        for i, attr in enumerate(self.basis):
            if self._mask[i]:
                if idx == id:
                    attr["id"] = idx
                    return attr
                else:
                    idx += 1

    def get_basis_info(self) -> Iterable[Dict]:
        idx = 0
        for i, attr in enumerate(self.basis):
            if self._mask[i]:
                attr["id"] = idx
                idx += 1
                yield attr

    def dimension(self) -> int:
        return self.sampled_basis.shape[0]
    
    def __eq__(self, other):
        if isinstance(other, SingleBlockBasisExpansion):
            return (
                    self.basis == other.basis and
                    torch.allclose(self.sampled_basis, other.sampled_basis) and
                    (self._mask == other._mask).all()
            )
        else:
            return False
    
    def __hash__(self):
        return 10000 * hash(self.basis) + 100 * hash(self.sampled_basis) + hash(self._mask)

# dictionary storing references to already built basis tensors.
# when a new filter tensor is built, it is also stored here
# when the same basis is built again (eg. in another layer), the already existing filter tensor is retrieved
_stored_filters = {}


def block_basisexpansion(basis: KernelBasis,
                         points: np.ndarray,
                         basis_filter: Callable[[dict], bool] = None,
                         recompute: bool = False
                         ) -> SingleBlockBasisExpansion:
    r"""


    Args:
        basis (KernelBasis): basis defining the space of kernels
        points (ndarray): points where the analytical basis should be sampled
        basis_filter (callable, optional): filter for the basis elements. Should take a dictionary containing an
                                           element's attributes and return whether to keep it or not.
        recompute (bool, optional): whether to recompute new bases or reuse, if possible, already built tensors.

    """

    if basis_filter is not None:
        mask = np.zeros(len(basis), dtype=bool)
        for b, attr in enumerate(basis):
            mask[b] = basis_filter(attr)
    else:
        mask = np.ones(len(basis), dtype=bool)

    if not recompute:
        # compute the mask of the sampled basis containing only the elements allowed by the filter
        key = (basis, mask.tobytes(), points.tobytes())
        if key not in _stored_filters:
            _stored_filters[key] = SingleBlockBasisExpansion(basis, points, mask)

        return _stored_filters[key]
    
    else:
        return SingleBlockBasisExpansion(basis, points, mask)


def normalize_basis(basis: torch.Tensor, sizes: torch.Tensor) -> torch.Tensor:
    r"""

    Normalize the filters in the input tensor.
    The tensor of shape :math:`(B, O, I, ...)` is interpreted as a basis containing ``B`` filters/elements, each with
    ``I`` inputs and ``O`` outputs. The spatial dimensions ``...`` can be anything.

    .. notice ::
        Notice that the method changes the input tensor inplace

    Args:
        basis (torch.Tensor): tensor containing the basis to normalize
        sizes (torch.Tensor): original input size of the basis elements, without the padding and the change of basis

    Returns:
        the normalized basis (the operation is done inplace, so this is ust a reference to the input tensor)

    """
    
    b = basis.shape[0]
    assert len(basis.shape) > 2, basis.shape
    assert sizes.shape == (b,), (sizes.shape, b, basis.shape)
    
    # compute the norm of each basis vector
    norms = torch.einsum('bop...,bpq...->boq...', (basis, basis.transpose(1, 2)))
    
    # Removing the change of basis, these matrices should be multiples of the identity
    # where the scalar on the diagonal is the variance
    # in order to find this variance, we can compute the trace (which is invariant to the change of basis)
    # and divide by the number of elements in the diagonal ignoring the padding.
    # Therefore, we need to know the original size of each basis element.
    norms = torch.einsum("bii...->b", norms)
    # norms = norms.reshape(b, -1).sum(1)
    norms /= sizes

    norms[norms < 1e-15] = 0
    
    norms = torch.sqrt(norms)
    
    norms[norms < 1e-6] = 1
    norms[norms != norms] = 1
    
    norms = norms.view(b, *([1] * (len(basis.shape) - 1)))
    
    # divide by the norm
    basis /= norms

    return basis







from .basis import KernelBasis, EmptyBasisException

from escnn.group import Group
from escnn.group import IrreducibleRepresentation
from escnn.group import Representation

import torch

from typing import Type, Union, Tuple, Dict, List, Iterable, Callable, Set
from abc import ABC, abstractmethod
from collections import defaultdict

import numpy as np


class SteerableFiltersBasis(KernelBasis):
    
    def __init__(self,
                 G: Group,
                 action: Representation,
                 js: List[Tuple],
        ):
        r"""

        Abstract class for bases implementing a :math:`G`-steerable basis for *scalar* filters over an Euclidean space
        :math:`\R^n`.

        The action of ``G`` on the Euclidean space is given by the :class:`~escnn.group.Representation` ``action``.
        The dimensionality of the Euclidean space is inferred by the size of the ``action``.

        A :math:`G`-steerable basis provides an irreps-decomposition of the action of :math:`G` on the vector space of
        square-integrable functions, i.e. :math:`L^2(\R^n)`.
        The input list ``js`` defines the order and the multiplicity of the ``G``-irreps in this decomposition.
        More precisely, ``js`` is a list of tuples ``(irrep_id, multiplicity)``, where ``irrep_id`` is the ``id`` of
        an :class:`~escnn.group.IrreducibleRepresentation`.

        The order of the basis elements sampled in :meth:`~escnn.kernels.SteerableFiltersBasis.sample` should be
        consistent with the order of the irreps in the list ``js``.

        Since this class only parameterizes scalar filters, ``SteerableFiltersBasis.shape = (1, 1)``.

        Args:
            G (Group): the symmetry group
            action (Representation): the representation of the action of ``G`` on the Euclidean space
            js (list): the multiplicity of each irrep in this basis.

        Attributes:
            ~.G (Group): the symmetry group
            ~.action (Representation): the representation of the action of ``G`` on the Euclidean space
            ~.js (list): the multiplicity of each irrep in this basis.

        """
        assert isinstance(G, Group)

        # Group: the group acting on the steerable basis
        self.group: Group = G

        # Representation: the representation of the group ``G`` defining its action on the Euclidean space
        assert action.group == self.group
        self.action = action

        assert isinstance(js, list)
        # List: list of irreps (and their multiplicity) describing how each invariant steerable subspace transform
        self.js = js

        self._js = {}
        dim = 0
        for j, m in js:
            assert isinstance(j, tuple)
            # check it corresponds to an irrep
            psi_j = self.group.irrep(*j)

            # the second entry represents the multiplicity of the irrep
            assert isinstance(m, int)
            self._js[j] = m

            dim += psi_j.size * m

        # This is a Filter basis, so it assumes 1 input and 1 output channels
        super(SteerableFiltersBasis, self).__init__(dim, (1, 1))

        self._start_index = {}
        idx = 0
        for _j, m in self.js:
            self._start_index[_j] = idx
            idx += self.dim_harmonic(_j)

    @property
    def dimensionality(self) -> int:
        """
            The dimensionality of the Euclidean space on which the scalar filters are defined.
        """
        return self.action.size

    @abstractmethod
    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.
        Optionally, store the resulting multidimensional array in ``out``.

        ``points`` must be an array of shape `(N, d)`, where `N` is the number of points and `d` is equal to
        :meth:`~escnn.kernels.SteerableFilterBasis.dimensionality`.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        raise NotImplementedError
        # assert len(points.shape) == 2
        # S = points.shape[1]
        # assert points.shape[0] == self.dimensionality, (points.shape, self.dimensionality)
        #
        # if out is None:
        #     out = torch.empty(1, 1, self.dim, S, device=points.device, dtype=points.dtype)
        #
        # assert out.shape == (1, 1, self.dim, S)
        #
        # B = 0
        # for b, j in enumerate(self.js):
        #     self.sample_j(points, j, out=out[:, :, B:B + self.dim_harmonic(j), :])
        #     B += self.dim_harmonic(j)
        #
        # return out

    def sample_as_dict(self, points: torch.Tensor, out: torch.Tensor = None) -> Dict[Tuple, torch.Tensor]:
        r"""

        Sample the continuous basis elements on the discrete set of points in ``points``.

        Rather than returning a single tensor containing all sampled basis elements, it groups basis elements by
        the ``G``-irrep acting on them.
        Then, the method returns a dictionary mapping each irrep's ``id`` to a tensor of shape `(N, m, d)`, where
        `m` is the multiplicity of the irrep (as in the list ``js``) and `d` is the size of the irrep.

        ``points`` must be an array of shape `(N, D)`, where `N` is the number of points and `D` is equal to
        :meth:`~escnn.kernels.SteerableFilterBasis.dimensionality`.

        Optionally, store the resulting multidimentional array in ``out``.

        Args:
            points (~torch.Tensor): points where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """

        S = points.shape[0]

        if out is not None:
            assert out.shape == (S, self.dim), (out.shape, self.dim, S)
            out = out.view(S, self.dim, 1, 1)

        out = self.sample(points, out)

        out = out.view(S, self.dim)

        basis = {}
        p = 0
        for j, m in self.js:
            psi = self.group.irrep(*j)
            dim = psi.size * m
            basis[j] = out[:, p : p+dim].view(S, m, psi.size)
            p += dim

        return basis

    def dim_harmonic(self, j: Tuple) -> int:
        r"""
            Number of basis elements associated with the ``G``-irrep with the id ``j``.
            This is equal to the multiplicity of this irrep times its size, i.e. the number of subspaces transforming
            according to this irrep times the dimensionality of these subspaces.
        """
        psi = self.group.irrep(*j)
        if j in self._js:
            return psi.size * self._js[j]
        else:
            return 0

    def multiplicity(self, j: Tuple) -> int:
        r"""
            The multiplicity of the ``G``-irrep with the id ``j`` as defined in the list ``js``.
        """
        if j in self._js:
            return self._js[j]
        else:
            return 0

    def _get_j_from_idx(self, idx: int):
        assert idx < self.dim
        p = 0
        for j, m in self.js:
            dim = self.dim_harmonic(j)
            if p + dim > idx:
                return j, idx - p
            p += dim
        raise ValueError()

    def _get_j_from_idx_steerable(self, idx: int):
        p = 0
        for j, m in self.js:
            if p + m > idx:
                return j, idx - p
            p += m
        raise ValueError()

    # def __getitem__(self, idx):
    #     j, rel_idx = self._get_j_from_idx(idx)
    #     return self.attrs_j(j, rel_idx)
    #
    # def __iter__(self):
    #     for j, _ in self.js:
    #         for attr in self.attrs_j_iter(j):
    #             yield attr

    def steerable_attrs_iter(self) -> Iterable:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis
        for j, _ in self.js:
            for attr in self.steerable_attrs_j_iter(j):
                yield attr

    def steerable_attrs(self, idx) -> Dict:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis
        j, rel_idx = self._get_j_from_idx_steerable(idx)
        return self.steerable_attrs_j(j, rel_idx)

    @abstractmethod
    def steerable_attrs_j_iter(self, j: Tuple) -> Iterable:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis
        raise NotImplementedError()

    @abstractmethod
    def steerable_attrs_j(self, j: Tuple, idx) -> Dict:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis
        raise NotImplementedError()

    def check_equivariance(self):
        # Verify the steerability property of the basis

        S = 20
        points = torch.randn(S, self.dimensionality)

        basis = self.sample_as_dict(points)

        for _ in range(10):
            g = self.group.sample()

            points_g = points @ torch.tensor(self.action(g), device=points.device, dtype=points.dtype).T
            basis_g = self.sample_as_dict(points_g)

            g_basis = {}
            for j, basis_j in basis.items():
                rho_g = torch.tensor(self.group.irrep(*j)(g), device=points.device, dtype=points.dtype)
                g_basis[j] = torch.einsum(
                    'ij,pmj->pmi',
                    rho_g,
                    basis_j,
                )

            for j, m in self.js:
                dim = self.group.irrep(*j).size
                assert basis_g[j].shape == (S, m, dim), (basis_g[j].shape, m, dim, S)
                assert g_basis[j].shape == (S, m, dim), (g_basis[j].shape, m, dim, S)
                assert torch.allclose(g_basis[j], basis_g[j], atol=2e-6, rtol=5e-4), (g_basis[j]-basis_g[j]).max().item()


class PointBasis(SteerableFiltersBasis):

    def __init__(self, G: Group):
        r"""
            Basis for scalar functions over a "0-dimensional" Euclidean space, i.e. a space containing only a single
            point.
            This basis only contains the function associating the value :math:`1` to this single point.

            This basis is steerable for any group ``G``, which acts on this point by mapping it to itself.

            This class is useful to parameterize equivairant linear layers as a special case of steerable convolution
            layers in a neural network.

            .. warning::
                This class is implemented by using ``G.trivial_representation`` as ``action`` which, however, implies
                a 1-dimensional Euclidean space.

        """
        super(PointBasis, self).__init__(G, G.trivial_representation, [(G.trivial_representation.id, 1)])

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:

        assert len(points.shape) == 2
        assert points.shape[1] == self.dimensionality, (points.shape, self.dimensionality)

        S = points.shape[0]

        if out is None:
            return torch.ones(S, self.dim, 1, 1, device=points.device, dtype=points.dtype)
        else:
            assert out.shape == (S, self.dim, 1, 1)
            out[:] = 1.
            return out

    def steerable_attrs_iter(self):
        yield self.steerable_attrs(0)

    def steerable_attrs(self, idx):
        assert idx == 0, idx
        attr = {}
        attr["idx"] = 0
        attr["j"] = self.group.trivial_representation.id
        attr["shape"] = (1, 1)
        return attr

    def steerable_attrs_j_iter(self, j: Tuple) -> Iterable:
        if j != self.group.trivial_representation.id:
            return
        yield self.steerable_attrs(0)

    def steerable_attrs_j(self, j: Tuple, idx) -> Dict:
        if j != self.group.trivial_representation.id:
            return
        return self.steerable_attrs(idx)

    def __getitem__(self, idx):
        return self.steerable_attrs(idx)

    def __iter__(self):
        yield self.steerable_attrs(0)

    def __eq__(self, other):
        if isinstance(other, PointBasis):
            return self.group == other.group
        else:
            return False

    def __hash__(self):
        return hash(self.group.__class__) * 1000 + sum(hash(x) for x in self.group._keys.items())




import numpy as np
import torch

from escnn.group import *

from abc import ABC, abstractmethod
from typing import List, Union, Tuple, Callable, Dict, Iterable

from .basis import KernelBasis, EmptyBasisException
from .steerable_filters_basis import SteerableFiltersBasis

from .harmonic_polynomial_r3 import HarmonicPolynomialR3Generator


__all__ = [
    'GaussianRadialProfile',
    'SphericalShellsBasis'
]


class GaussianRadialProfile(KernelBasis):
    
    def __init__(self, radii: List[float], sigma: Union[List[float], float]):
        r"""

        Basis for kernels defined over a radius in :math:`\R^+_0`.

        Each basis element is defined as a Gaussian function.
        Different basis elements are centered at different radii (``rings``) and can possibly be associated with
        different widths (``sigma``).

        More precisely, the following basis is implemented:

        .. math::

            \mathcal{B} = \left\{ b_i (r) :=  \exp \left( \frac{ \left( r - r_i \right)^2}{2 \sigma_i^2} \right) \right\}_i

        In order to build a complete basis of kernels, you should combine this basis with a basis which defines the
        angular profile, see for example :class:`~escnn.kernels.SphericalShellsBasis` or
        :class:`~escnn.kernels.CircularShellsBasis`.

        Args:
            radii (list): centers of each basis element. They should be different and spread to cover all
                domain of interest
            sigma (list or float): widths of each element. Can potentially be different.


        """
        
        if isinstance(sigma, float):
            sigma = [sigma] * len(radii)
        
        assert len(radii) == len(sigma)
        assert isinstance(radii, list)
        
        for r in radii:
            assert r >= 0.
        
        for s in sigma:
            assert s > 0.
        
        super(GaussianRadialProfile, self).__init__(len(radii), (1, 1))
        
        self.register_buffer('radii', torch.tensor(radii, dtype=torch.float32).reshape(1, -1, 1, 1))
        self.register_buffer('sigma', torch.tensor(sigma, dtype=torch.float32).reshape(1, -1, 1, 1))
    
    def sample(self, radii: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on the discrete set of radii in ``radii``.
        Optionally, store the resulting multidimentional array in ``out``.

        ``radii`` must be an array of shape `(N, 1)`, where `N` is the number of points.

        Args:
            radii (~torch.Tensor): radii where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(radii.shape) == 2
        assert radii.shape[1] == 1
        S = radii.shape[0]
        
        if out is None:
            out = torch.empty((S, self.dim, self.shape[0], self.shape[1]), device=radii.device, dtype=radii.dtype)
        
        assert out.shape == (S, self.dim, self.shape[0], self.shape[1])
        
        radii = radii.reshape(-1, 1, 1, 1)

        assert not torch.isnan(radii).any()

        d = (self.radii - radii) ** 2

        if radii.requires_grad:
            out[:] = torch.exp(-0.5 * d / self.sigma ** 2)
        else:
            out = torch.exp(-0.5 * d / self.sigma ** 2, out=out)

        return out
    
    def __getitem__(self, r):
        assert r < self.dim
        return {"radius": self.radii[0, r, 0, 0].item(), "sigma": self.sigma[0, r, 0, 0].item(), "idx": r}
    
    def __eq__(self, other):
        if isinstance(other, GaussianRadialProfile):
            return (
                        torch.allclose(self.radii, other.radii.to(self.radii.device))
                    and torch.allclose(self.sigma, other.sigma.to(self.sigma.device))
            )
        else:
            return False
    
    def __hash__(self):
        return hash(self.radii.cpu().numpy().tobytes()) + hash(self.sigma.cpu().numpy().tobytes())


def circular_harmonics(points: torch.Tensor, L: int, phase: float = 0.):
    r"""
        Compute the circular harmonics up to frequency ``L``.
    """

    assert len(points.shape) == 2
    assert points.shape[1] == 2

    device = points.device
    dtype = points.dtype

    S = points.shape[0]

    x, y = points.T

    angles = torch.atan2(y, x).view(S, 1) - phase

    freqs = torch.arange(1, L+1, device=device, dtype=dtype).view(1, L)

    freqs_times_angles = freqs * angles

    del freqs, angles

    Y = torch.empty((S, 2 * L + 1), dtype=dtype, device=device)

    Y[:, 0] = 1.
    Y[:, 1::2] = torch.cos(freqs_times_angles)
    Y[:, 2::2] = torch.sin(freqs_times_angles)

    return Y


class SphericalShellsBasis(SteerableFiltersBasis):

    def __init__(self,
                 L: int,
                 radial: GaussianRadialProfile,
                 filter: Callable[[Dict], bool] = None
                 ):
        r"""

        Build the tensor product basis of a radial profile basis and a spherical harmonics basis for kernels over the
        Euclidean space :math:`\R^3`.
        
        The kernel space is spanned by an independent basis for each shell.
        The kernel space over each shell is spanned by the spherical harmonics of frequency up to `L`
        (an independent copy of each for each cell).

        Given the bases :math:`A = \{a_j\}_j` for the spherical shells and
        :math:`D = \{d_r\}_r` for the radial component (indexed by :math:`r \geq 0`, the radius of each ring),
        this basis is defined as

        .. math::
            C = \left\{c_{i,j}(\bold{p}) := d_r(||\bold{p}||) a_j(\hat{\bold{p}}) \right\}_{r, j}

        where :math:`(||\bold{p}||, \hat{\bold{p}})` are the polar coordinates of the point
        :math:`\bold{p} \in \R^n`.
        
        The radial component is parametrized using :class:`~escnn.kernels.GaussianRadialProfile`.
        
        Args:
            L (int): the maximum spherical frequency
            radial (GaussianRadialProfile): the basis for the radial profile
            filter (callable, optional): function used to filter out some basis elements. It takes as input a dict
                describing a basis element and should return a boolean value indicating whether to keep (`True`) or
                discard (`False`) the element. By default (`None`), all basis elements are kept.

        Attributes:
            ~.radial (GaussianRadialProfile): the radial basis
            ~.L (int): the maximum spherical frequency

        """

        self.L: int = L

        assert isinstance(radial, GaussianRadialProfile)

        self._angular_dim = (L+1)**2

        # number of invariant subspaces
        self._num_inv_spaces = 0

        G = o3_group(L)

        if filter is not None:
            _filter = torch.zeros(self._angular_dim * len(radial), dtype=torch.bool)

            js = []
            _idx_map = []
            _steerable_idx_map = []
            i = 0
            steerable_i = 0
            for j in range(self.L+1):

                j_id = (j % 2, j)  # the id of the O(3) irrep
                attr2 = {
                    'irrep:' + k: v
                    for k, v in G.irrep(*j_id).attributes.items()
                }
                attr2['j'] = j_id
                dim = 2 * j + 1

                multiplicity = 0

                for attr1 in radial:
                    attr = dict()
                    attr.update(attr1)
                    attr.update(attr2)

                    if filter(attr):
                        multiplicity += 1
                        _filter[i:i+dim] = 1
                        _idx_map += list(range(i, i+dim))
                        _steerable_idx_map.append(steerable_i)

                    i += dim
                    steerable_i += 1

                js.append(
                    (
                        (j%2, j), # the O(3) irrep ID
                        multiplicity
                    )
                )
                self._num_inv_spaces += multiplicity
                    
            self._idx_map = np.array(_idx_map)
            self._steerable_idx_map = np.array(_steerable_idx_map)
        else:
            _filter = None
            self._idx_map = None
            self._steerable_idx_map = None
            js = [
                (
                    (j % 2, j),  # the O(3) irrep ID
                    len(radial)
                )
                for j in range(L+1)
            ]

        super(SphericalShellsBasis, self).__init__(G, G.standard_representation(), js)

        self.radial = radial

        self.harmonics_generator = HarmonicPolynomialR3Generator(self.L)

        if _filter is not None:
            self.register_buffer('_filter', _filter)
        else:
            self._filter = None

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on a discrete set of ``points`` in the space :math:`\R^n`.
        Optionally, store the resulting multidimensional array in ``out``.

        ``points`` must be an array of shape `(N, 3)` containing `N` points in the space.
        Note that the points are specified in cartesian coordinates :math:`(x, y, z, ...)`.

        Args:
            points (~torch.Tensor): points in the n-dimensional Euclidean space where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(points.shape) == 2
        assert points.shape[1] == self.dimensionality, (points.shape, self.dimensionality)

        S = points.shape[0]

        assert not torch.isnan(points).any()

        radii = torch.norm(points, dim=1, keepdim=True)

        non_origin_mask = (radii > 1e-9).reshape(-1)
        any_origin = not non_origin_mask.all()
        # sphere = points[non_origin_mask, :] / radii[non_origin_mask, :]
        if any_origin:
            sphere = torch.empty_like(points)
            sphere[non_origin_mask, :] = torch.nn.functional.normalize(points[non_origin_mask], dim=1)
            sphere[~non_origin_mask, :] = points[~non_origin_mask]
        else:
            sphere = torch.nn.functional.normalize(points, dim=1)

        if out is None:
            out = torch.empty(S, self.dim, 1, 1, device=points.device, dtype=points.dtype)
        
        assert out.shape == (S, self.dim, 1, 1)
        
        # sample the radial basis
        radial = self.radial.sample(radii)
        assert radial.shape == (S, len(self.radial), 1, 1)
        radial = radial[..., 0, 0]

        assert not torch.isnan(radial).any()

        # sample the angular basis
        spherical = self.harmonics_generator(sphere)

        # only frequency 0 is sampled at the origin. Other frequencies are set to 0
        # spherical[~non_origin_mask, :1] = 1.
        # spherical[~non_origin_mask, 1:] = 0.
        if any_origin:
            assert (spherical[~non_origin_mask, :1]-1.).abs().max().item() < 1e-7, (spherical[~non_origin_mask, :1]-1.).abs().max().item()
            if spherical.shape[1] > 1:
                assert (spherical[~non_origin_mask, 1:]).abs().max().item() < 1e-7, (spherical[~non_origin_mask, 1:]).abs().max().item()

        assert not torch.isnan(spherical).any()

        tensor_product = torch.einsum("pa,pb->pab", radial, spherical)

        n_radii = len(self.radial)

        if self._filter is None:
            tmp_out = out
        else:
            tmp_out = torch.empty(S, self._angular_dim*n_radii, 1, 1, device=points.device, dtype=points.dtype)

        for j in range(self.L+1):
            first, last = j**2, (j+1)**2
            tmp_out[:, first * n_radii:last * n_radii, 0, 0].view(
                S, n_radii, 2*j+1
            )[:] = tensor_product[:, :, first:last]

        if self._filter is not None:
            out[:] = tmp_out[:, self._filter, ...]

        return out

    def steerable_attrs_iter(self):
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        idx = 0
        i = 0

        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)

        for j in range(self.L + 1):
            attr1 = {
                'irrep:' + k: v
                for k, v in self.group.irrep(j%2, j).attributes.items()
            }

            for radial_idx, attr2 in enumerate(radial_attrs):
                if self._filter is None or (self._filter[i:i+2*j+1] == 1).all():
                    assert attr2["idx"] == radial_idx

                    attr = dict()
                    attr.update(attr1)
                    attr.update(attr2)
                    attr["idx"] = idx
                    attr["radial_idx"] = radial_idx
                    attr["j"] = (j % 2, j)  # the id of the O(3) irrep
                    attr["shape"] = (1, 1)

                    yield attr
                    idx += 1
                i += 2*j+1

    def steerable_attrs(self, idx):
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        assert idx < self._num_inv_spaces, (idx, self._num_inv_spaces)

        if self._steerable_idx_map is None:
            _idx = idx
        else:
            _idx = self._steerable_idx_map[idx]

        j, radial_idx = divmod(_idx, len(self.radial))

        attr2 = self.radial[radial_idx]

        assert attr2["idx"] == radial_idx

        attr = dict()
        attr.update(attr2)

        j = (j%2, j) # the id of the O(3) irrep
        attr = {
            'irrep:'+k: v
            for k, v in self.group.irrep(*j).attributes.items()
        }
        attr["j"] = j

        attr["idx"] = idx
        attr["radial_idx"] = radial_idx
        attr["shape"] = (1, 1)

        return attr

    def steerable_attrs_j_iter(self, j: Tuple) -> Iterable:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        j_id = j
        f, j = j_id

        if f != j%2:
            return

        idx = sum(self.multiplicity((_j%2, _j)) for _j in range(j))
        i = 0

        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)

        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }

        for radial_idx, attr2 in enumerate(radial_attrs):
            if self._filter is None or (self._filter[i:i+2*j+1] == 1).all():
                assert attr2["idx"] == radial_idx

                attr = dict()
                attr.update(attr1)
                attr.update(attr2)
                attr["idx"] = idx
                attr["radial_idx"] = radial_idx
                attr["j"] = j_id
                attr["shape"] = (1, 1)

                yield attr
                idx += 1
            i += 2*j+1

    def steerable_attrs_j(self, j: Tuple, idx) -> Dict:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        j_id = j
        f, j = j_id

        if f != j % 2:
            return

        assert idx < self.multiplicity(j_id), (idx, self.multiplicity(j_id))

        idx += sum(self.multiplicity((_j%2, _j)) for _j in range(j))

        if self._steerable_idx_map is None:
            _idx = idx
        else:
            _idx = self._steerable_idx_map[idx]

        _j, radial_idx = divmod(_idx, len(self.radial))
        assert _j == j, (j, _j)

        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }

        attr2 = self.radial[radial_idx]

        assert attr2["idx"] == radial_idx

        attr = dict()
        attr.update(attr1)
        attr.update(attr2)

        attr["idx"] = idx
        attr["radial_idx"] = radial_idx
        attr["j"] = j_id
        attr["shape"] = (1, 1)

        return attr

    def __getitem__(self, idx):
        assert idx < self.dim, (idx, self.dim)
        
        if self._idx_map is None:
            _idx = idx
        else:
            _idx = self._idx_map[idx]

        j = int(np.floor(np.sqrt(_idx // len(self.radial))))

        assert j**2 * len(self.radial) <= _idx < (j+1)**2 * len(self.radial), (_idx, j, self.L, len(self.radial))

        j_idx = _idx - j**2 * len(self.radial)
        
        radial_idx, m = divmod(j_idx, 2*j+1)

        j_id = (j%2, j) # the id of the O(3) irrep
        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }

        attr2 = self.radial[radial_idx]

        assert attr2["idx"] == radial_idx

        attr = dict()
        attr.update(attr1)
        attr.update(attr2)

        attr["idx"] = idx
        attr["radial_idx"] = radial_idx
        attr["j"] = j_id
        attr["m"] = m
        attr["shape"] = (1, 1)

        return attr

    def __iter__(self):
        idx = 0
        i = 0

        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)

        for j in range(self.L+1):

            j_id = (j % 2, j)  # the id of the O(3) irrep
            attr1 = {
                'irrep:' + k: v
                for k, v in self.group.irrep(*j_id).attributes.items()
            }
            for radial_idx, attr2 in enumerate(radial_attrs):
                for m in range(2*j+1):
                    if self._filter is None or self._filter[i] == 1:
                        assert attr2["idx"] == radial_idx

                        attr = dict()
                        attr.update(attr1)
                        attr.update(attr2)
                        attr["idx"] = idx
                        attr["radial_idx"] = radial_idx
                        attr["j"] = j_id
                        attr["m"] = m
                        attr["shape"] = (1, 1)

                        yield attr
                        idx += 1
                    i += 1
    
    def __eq__(self, other):
        if isinstance(other, SphericalShellsBasis):
            return (
                    self.radial == other.radial and
                    self.L == other.L and
                    self._filter == other._filter
            )
        else:
            return False
    
    def __hash__(self):
        return self.L + hash(self.radial) + hash(self._filter)


class CircularShellsBasis(SteerableFiltersBasis):

    def __init__(self,
                 L: int,
                 radial: GaussianRadialProfile,
                 filter: Callable[[Dict], bool] = None,
                 axis: float = np.pi/2,
                 ):
        r"""

        Build the tensor product basis of a radial profile basis and a circular harmonics basis for kernels over the
        Euclidean space :math:`\R^2`.

        The kernel space is spanned by an independent basis for each shell.
        The kernel space each shell is spanned by the circular harmonics of frequency up to `L`
        (an independent copy of each for each cell).

        Given the bases :math:`A = \{a_j\}_j` for the circular shells and
        :math:`D = \{d_r\}_r` for the radial component (indexed by :math:`r \geq 0`, the radius different rings),
        this basis is defined as

        .. math::
            C = \left\{c_{i,j}(\bold{p}) := d_r(||\bold{p}||) a_j(\hat{\bold{p}}) \right\}_{r, j}

        where :math:`(||\bold{p}||, \hat{\bold{p}})` are the polar coordinates of the point
        :math:`\bold{p} \in \R^n`.

        The radial component is parametrized using :class:`~escnn.kernels.GaussianRadialProfile`.

        Args:
            L (int): the maximum circular frequency
            radial (GaussianRadialProfile): the basis for the radial profile
            filter (callable, optional): function used to filter out some basis elements. It takes as input a dict
                describing a basis element and should return a boolean value indicating whether to keep (`True`) or
                discard (`False`) the element. By default (`None`), all basis elements are kept.

        Attributes:
            ~.radial (GaussianRadialProfile): the radial basis
            ~.L (int): the maximum circular frequency

        """

        self.L: int = L

        assert isinstance(radial, GaussianRadialProfile)

        self._angular_dim = 2*L+1

        # number of invariant subspaces
        self._num_inv_spaces = 0

        G = o2_group(L)

        if filter is not None:
            _filter = torch.zeros(self._angular_dim * len(radial), dtype=torch.bool)

            js = []
            _idx_map = []
            _steerable_idx_map = []
            i = 0
            steerable_i = 0
            for j in range(self.L + 1):

                attr2 = {
                    'irrep:' + k: v
                    for k, v in G.irrep(int(j>0), j).attributes.items()
                }
                attr2['j'] = (int(j>0), j)  # the id of the O(2) irrep

                dim = 2 if j > 0 else 1

                multiplicity = 0

                for attr1 in radial:
                    attr = dict()
                    attr.update(attr1)
                    attr.update(attr2)

                    if filter(attr):
                        multiplicity += 1
                        _filter[i:i + dim] = 1
                        _idx_map += list(range(i, i + dim))
                        _steerable_idx_map.append(steerable_i)

                    i += dim
                    steerable_i += 1

                js.append(
                    (
                        (int(j>0), j),  # the O(2) irrep ID
                        multiplicity
                    )
                )
                self._num_inv_spaces += multiplicity

            self._idx_map = np.array(_idx_map)
            self._steerable_idx_map = np.array(_steerable_idx_map)
        else:
            _filter = None
            self._idx_map = None
            self._steerable_idx_map = None
            js = [
                (
                    (int(j>0), j),  # the O(2) irrep ID
                    len(radial)
                )
                for j in range(L + 1)
            ]

        self.axis = axis

        action = G.standard_representation()
        action = change_basis(action, action(G.element((0, axis), 'radians')), name=f'StandardAction|axis=[{axis}]')
        super(CircularShellsBasis, self).__init__(G, action, js)

        self.radial = radial

        if _filter is None:
            self._filter = None
        else:
            self.register_buffer('_filter', _filter)

    def sample(self, points: torch.Tensor, out: torch.Tensor = None) -> torch.Tensor:
        r"""

        Sample the continuous basis elements on a discrete set of ``points`` in the space :math:`\R^n`.
        Optionally, store the resulting multidimensional array in ``out``.

        ``points`` must be an array of shape `(N, 2)` containing `N` points in the space.
        Note that the points are specified in cartesian coordinates :math:`(x, y, z, ...)`.

        Args:
            points (~torch.Tensor): points in the n-dimensional Euclidean space where to evaluate the basis elements
            out (~torch.Tensor, optional): pre-existing array to use to store the output

        Returns:
            the sampled basis

        """
        assert len(points.shape) == 2
        assert points.shape[1] == self.dimensionality, (points.shape, self.dimensionality)

        S = points.shape[0]

        radii = torch.norm(points, dim=1, keepdim=True)

        non_origin_mask = (radii > 1e-9).reshape(-1)
        sphere = points[non_origin_mask, :] / radii[non_origin_mask, :]

        if out is None:
            out = torch.empty(S, self.dim, 1, 1, device=points.device, dtype=points.dtype)

        assert out.shape == (S, self.dim, 1, 1)

        # sample the radial basis
        radial = self.radial.sample(radii)
        assert radial.shape[-2:] == (1, 1)
        radial = radial[..., 0, 0]

        assert not torch.isnan(radial).any()

        # sample the angular basis
        circular = torch.empty(S, self._angular_dim, device=points.device, dtype=points.dtype)
        # circular[:] = np.nan

        # where r>0, we sample all frequencies
        circular[non_origin_mask, :] = circular_harmonics(sphere, self.L, phase=self.axis)

        # only frequency 0 is sampled at the origin. Other frequencies are set to 0
        circular[~non_origin_mask, :1] = 1.

        # This trick allows us to compute meaningful gradients at the origin
        # Unfortunately, only the newest versions of PyTorch and CUDA support these complex operations
        # complex_points = points[~non_origin_mask, 0] + 1j * points[~non_origin_mask, 1]
        # complex_powers = complex_points.view(S, 1).pow(torch.arange(1, self.L).view(1, self.L))
        # circular[~non_origin_mask, 1::2] = complex_powers.real
        # circular[~non_origin_mask, 2::2] = complex_powers.img
        circular[~non_origin_mask, 1:] = 0.

        tensor_product = torch.einsum("pa,pb->pab", radial, circular)

        n_radii = len(self.radial)

        if self._filter is None:
            tmp_out = out
        else:
            tmp_out = torch.empty(S, self._angular_dim*n_radii, 1, 1, device=points.device, dtype=points.dtype)

        for j in range(self.L+1):
            dim = 2 if j > 0 else 1
            last = 2*j+1
            first = last - dim
            tmp_out[:, first * n_radii:last * n_radii, 0, 0].view(
                S, n_radii, dim
            )[:] = tensor_product[:, :, first:last]

        if self._filter is not None:
            out[:] = tmp_out[:, self._filter, ...]

        return out

    def steerable_attrs_iter(self):
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        idx = 0
        i = 0

        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)

        for j in range(self.L + 1):
            dim = 2 if j > 0 else 1
            j_id = (int(j>0), j)  # the id of the O(2) irrep

            attr1 = {
                'irrep:' + k: v
                for k, v in self.group.irrep(*j_id).attributes.items()
            }

            for radial_idx, attr2 in enumerate(radial_attrs):
                if self._filter is None or (self._filter[i:i + dim] == 1).all():
                    assert attr2["idx"] == radial_idx

                    attr = dict()
                    attr.update(attr1)
                    attr.update(attr2)
                    attr["idx"] = idx
                    attr["radial_idx"] = radial_idx
                    attr["j"] = j_id
                    attr["shape"] = (1, 1)

                    yield attr
                    idx += 1
                i += dim

    def steerable_attrs(self, idx):
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        assert idx < self._num_inv_spaces, (idx, self._num_inv_spaces)

        if self._steerable_idx_map is None:
            _idx = idx
        else:
            _idx = self._steerable_idx_map[idx]

        j, radial_idx = divmod(_idx, len(self.radial))

        j_id = (int(j > 0), j)  # the id of the O(2) irrep
        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }

        attr2 = self.radial[radial_idx]

        assert attr2["idx"] == radial_idx

        attr = dict()
        attr.update(attr1)
        attr.update(attr2)

        attr["idx"] = idx
        attr["radial_idx"] = radial_idx
        attr["j"] = j_id
        attr["shape"] = (1, 1)

        return attr

    def steerable_attrs_j_iter(self, j: Tuple) -> Iterable:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        j_id = j
        f, j = j

        if f != int(j>0):
            return

        idx = sum(self.multiplicity((int(_j>0),_j)) for _j in range(j))
        dim = 2 if j > 0 else 1
        i = 0

        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }
        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)

        for radial_idx, attr2 in enumerate(radial_attrs):
            if self._filter is None or (self._filter[i:i + dim] == 1).all():
                assert attr2["idx"] == radial_idx

                attr = dict()
                attr.update(attr1)
                attr.update(attr2)
                attr["idx"] = idx
                attr["radial_idx"] = radial_idx
                attr["j"] = j_id
                attr["shape"] = (1, 1)

                yield attr
                idx += 1
            i += dim

    def steerable_attrs_j(self, j: Tuple, idx) -> Dict:
        # This attributes don't describe a single basis element but a group of basis elements which span an invariant
        # subspace. This is needed to generate the attributes of the SteerableKernelBasis

        j_id = j
        f, j = j_id

        if f != int(j>0):
            return

        assert idx < self.multiplicity(j_id), (idx, self.multiplicity(j_id))

        idx += sum(self.multiplicity((int(_j>0), _j)) for _j in range(j))

        if self._steerable_idx_map is None:
            _idx = idx
        else:
            _idx = self._steerable_idx_map[idx]

        _j, radial_idx = divmod(_idx, len(self.radial))
        assert _j == j, (j, _j)

        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }

        attr2 = self.radial[radial_idx]

        assert attr2["idx"] == radial_idx

        attr = dict()
        attr.update(attr1)
        attr.update(attr2)

        attr["idx"] = idx
        attr["radial_idx"] = radial_idx
        attr["j"] = j_id
        attr["shape"] = (1, 1)

        return attr

    def __getitem__(self, idx):
        assert idx < self.dim, (idx, self.dim)

        if self._idx_map is None:
            _idx = idx
        else:
            _idx = self._idx_map[idx]

        j = (_idx // len(self.radial) + 1) //2

        assert (2*j+1) * len(self.radial) <= _idx < (2*j + 3) * len(self.radial), (_idx, j, self.L, len(self.radial))

        j_idx = _idx - (2*j +1) * len(self.radial)

        radial_idx, m = divmod(j_idx, 2 if j>0 else 1)

        j_id = (int(j>0), j)  # the id of the O(3) irrep
        attr1 = {
            'irrep:' + k: v
            for k, v in self.group.irrep(*j_id).attributes.items()
        }

        attr2 = self.radial[radial_idx]

        assert attr2["idx"] == radial_idx

        attr = dict()
        attr.update(attr1)
        attr.update(attr2)

        attr["idx"] = idx
        attr["radial_idx"] = radial_idx
        attr["j"] = j_id
        attr["m"] = m
        attr["shape"] = (1, 1)

        return attr

    def __iter__(self):
        idx = 0
        i = 0

        # since this methods return iterables of attributes built on the fly, load all attributes first and then
        # iterate on these lists
        radial_attrs = list(self.radial)

        for j in range(self.L + 1):
            dim = 2 if j>0 else 1
            j_id = (int(j>0), j)
            attr1 = {
                'irrep:' + k: v
                for k, v in self.group.irrep(*j_id).attributes.items()
            }
            for radial_idx, attr2 in enumerate(radial_attrs):
                for m in range(dim):
                    if self._filter is None or self._filter[i] == 1:
                        assert attr2["idx"] == radial_idx

                        attr = dict()
                        attr.update(attr1)
                        attr.update(attr2)
                        attr["idx"] = idx
                        attr["radial_idx"] = radial_idx
                        attr["j"] = j_id
                        attr["m"] = m
                        attr["shape"] = (1, 1)

                        yield attr
                        idx += 1
                    i += 1

    def __eq__(self, other):
        if isinstance(other, CircularShellsBasis):
            return (
                    self.radial == other.radial and
                    self.L == other.L and
                    self._filter == other._filter
            )
        else:
            return False

    def __hash__(self):
        return self.L + hash(self.radial) + hash(self._filter)




from setuptools import find_packages, setup

about = {}
with open("escnn/__about__.py") as fp:
    exec(fp.read(), about)

install_requires = [
    "torch>=1.3",
    "numpy",
    "scipy",
    "lie_learn",
    "joblib",
    "pymanopt",
    "autograd",
    "py3nj",
]


setup_requires = [""]
tests_require = ["scikit-learn", "scikit-image"]

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

download_url = "https://github.com/QUVA-Lab/escnn/archive/v{}.tar.gz".format(
    about["__version__"]
)

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__summary__"],
    author=about["__author__"],
    author_email=about["__email__"],
    url=about["__url__"],
    download_url=download_url,
    license=about["__license__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["test", "test.*"]),
    python_requires=">=3.7",
    keywords=[
        "pytorch",
        "cnn",
        "convolutional-networks" "equivariant",
        "isometries",
    ],
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
)



from __future__ import annotations

import escnn.group

from .utils import *

from escnn.group import Group, GroupElement

from escnn.group import IrreducibleRepresentation
from escnn.group import Representation
from escnn.group import directsum
from escnn.group import utils

import numpy as np


from typing import Tuple, Callable, Iterable, List, Dict, Any, Union


__all__ = ["O2"]


class O2(Group):

    PARAM = 'radians'
    PARAMETRIZATIONS = [
        'radians',      # real in 0., 2pi/N, ... i*2pi/N, ...
        # 'C',            # point in the unit circle (i.e. cos and sin of 'radians')
        'MAT',  # 2x2 rotation matrix
    ]

    def __init__(self, maximum_frequency: int = 6):
        r"""
        Build an instance of the orthogonal group :math:`O(2)` which contains reflections and continuous planar
        rotations.
        
        Any group element is either a rotation :math:`r_{\theta}` by an angle :math:`\theta \in [0, 2\pi)` or a
        reflection :math:`f` followed by a rotation, i.e. :math:`r_{\theta}f`.
        Two reflections gives the identity :math:`f \cdot f = e` and a reflection commutes with a rotation by
        inverting it, i.e. :math:`r_\theta \cdot f = f \cdot r_{-\theta}`.
        A group element :math:`r_{\theta}f^j` is implemented as a pair :math:`(j, \theta)` with :math:`j \in \{0, 1\}`
        and :math:`\theta \in [0, 2\pi)`.
        
        .. note ::
        
            Since the group has infinitely many irreducible representations, it is not possible to build all of them.
            Each irrep is associated to one unique integer frequency and the parameter ``maximum_frequency`` specifies
            the maximum frequency of the irreps to build.
            New irreps (associated to higher frequencies) can be manually created by calling the method
            :meth:`~escnn.group.O2.irrep` (see the method's documentation).
        
        
        Subgroup Structure.
        
        A subgroup of :math:`O(2)` is identified by a tuple ``id`` :math:`(\theta, M)`.
        
        Here, :math:`M` can be either a positive integer indicating the number of rotations in the subgroup or
        :math:`-1`, indicating that the subgroup contains all continuous rotations.
        :math:`\theta` is either ``None`` or an angle in :math:`[0, \frac{2\pi}{M})`.
        If :math:`\theta` is ``None``, the subgroup does not contain any reflections.
        Otherwise, the subgroup contains the reflection :math:`r_{\theta}f` along the axis of the current group rotated
        by :math:`\frac{\theta}{2}`.
        
        Valid combinations are:
        
        - (``None``, :math:`M>0`): restrict to the cyclic subgroup :math:`C_M` generated by :math:`\langle r_{2\pi/M} \rangle`.
        
        - (``None``, :math:`-1`): restrict to the subgroup :math:`SO(2)` containing only the rotations
        
        - (:math:`\theta`, :math:`M>0`): restrict to the dihedral subgroup :math:`D_{M}` generated by :math:`\langle r_{2\pi/M}, r_{\theta} f \rangle`
        
        In particular:
        
        - (:math:`0.`, :math:`1`): restrict to the reflection group generated by :math:`\langle f \rangle`
        
        - (:math:`0.`, :math:`M`): restrict to the dihedral subgroup :math:`D_{M}` generated by :math:`\langle r_{2\pi/M}, f \rangle`
        
        - (``None``, :math:`1`): restrict to the cyclic subgroup of order 1 containing only the identity
        
        
        Args:
            maximum_frequency (int, optional): the maximum frequency to consider when building the irreps of the group
        
        Attributes:
            
            ~.reflection: the reflection element :math:`(j, \theta) = (1, 0.)`

        """
        
        assert (isinstance(maximum_frequency, int) and maximum_frequency >= 0)
        
        super(O2, self).__init__("O(2)", True, False)
        
        self.rotation_order = -1
        
        self._maximum_frequency = maximum_frequency
        
        self._identity = self.element((0, 0.))
        self.reflection = self.element((1, 0.))
        
        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        raise ValueError(f'{self.name} is a continuous groups and '
                         f'some of its generators are infinitesimal. '
                         f'This is not currently supported')

    @property
    def identity(self) -> GroupElement:
        return self._identity

    @property
    def elements(self) -> List[GroupElement]:
        return None

    # @property
    # def elements_names(self) -> List[str]:
    #     return None

    @property
    def _keys(self) -> Dict[str, Any]:
        return dict()

    @property
    def subgroup_trivial_id(self):
        return (None, 1)

    @property
    def subgroup_self_id(self):
        return (0., -1)

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################
    
    def _inverse(self, element: Tuple[int, float], param: str = PARAM) -> Tuple[int, float]:
        r"""
        Return the inverse element of the input element.
        Given the element :math:`r_\theta f^j` as a pair :math:`(j, \theta)`,
        the method returns :math:`r_{-\theta}` (as :math:`(0, -\theta)`) if :math:`f = 0` and
        :math:`r_\theta f` (as :math:`(1, \theta)`) otherwise.
        
        Args:
            element (tuple): a group element :math:`r_{\theta}f^j` a pair :math:`(j, \theta)`

        Returns:
            its inverse
            
        """
        element = self._change_param(element, p_from=param, p_to='radians')
        inverse = element[0], -element[1] * (-1 if element[0] else 1)
        return self._change_param(inverse, p_from='radians', p_to=param)

    def _combine(self, e1: Tuple[int, float], e2: Tuple[int, float], param: str = PARAM, param1: str = None, param2: str = None) -> Tuple[int, float]:
        r"""
        Return the combination of the two input elements.
        Given two input element :math:`r_\alpha f^a` and :math:`r_\beta f^b`, the method returns
        :math:`r_\alpha f^a \cdot r_\beta f^b`.

        Args:
            e1 (tuple): a group element :math:`r_\alpha f^a` as a pair :math:`(a, \alpha)`
            e2 (tuple): another element :math:`r_\beta f^b` as a pair :math:`(b, \beta)`

        Returns:
            their combination :math:`r_\alpha f^a \cdot r_\beta f^b`
        """
        if param1 is None:
            param1 = param
        if param2 is None:
            param2 = param

        e1 = self._change_param(e1, p_from=param1, p_to='radians')
        e2 = self._change_param(e2, p_from=param2, p_to='radians')
        product = (e1[0] + e2[0]) % 2, e1[1] + (-1 if e1[0] else 1)*e2[1]
        return self._change_param(product, p_from='radians', p_to=param)

    def _equal(self, e1: Tuple[int, float], e2: Tuple[int, float], param: str = PARAM, param1: str = None, param2: str = None) -> bool:
        r"""

        Check if the two input values corresponds to the same element.

        See :meth:`escnn.group.SO2.equal` for more details.

        Args:
            e1 (tuple): an element
            e2 (tuple): another element
            
        Returns:
            whether they are the same element

        """
        if param1 is None:
            param1 = param
        if param2 is None:
            param2 = param

        e1 = self._change_param(e1, p_from=param1, p_to='radians')
        e2 = self._change_param(e2, p_from=param2, p_to='radians')
        return e1[0] == e2[0] and utils.cycle_isclose(e1[1], e2[1], 2 * np.pi)

    def _hash_element(self, element: Tuple[int, float], param: str = PARAM):
        element = self._change_param(element, p_from=param, p_to='radians')
        return hash(tuple(
            np.around(np.array([element[0], np.cos(element[1]), np.sin(element[1])]), 5)
        ))

    def _repr_element(self, element: Tuple[int, float], param: str = PARAM):
        element = self._change_param(element, p_from=param, p_to='radians')
        return "({}, {})".format('+' if not element[0] else '-', element[1])

    def _is_element(self, element: Tuple[int, float], param: str = PARAM, verbose: bool = False) -> bool:
        element = self._change_param(element, p_from=param, p_to='radians')
        if isinstance(element, tuple) and len(element) == 2 and isinstance(element[0], int) and isinstance(element[1], float):
            return element[0] in {0, 1}
        else:
            return False

    def _change_param(self, element: Tuple[int, float], p_from: str, p_to: str):
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS

        flip, rotation = element

        # convert to radians
        if p_from == 'MAT':
            assert isinstance(rotation, np.ndarray)
            assert rotation.shape == (2, 2)
            assert np.isclose(np.linalg.det(rotation), 1.)
            assert np.allclose(rotation @ rotation.T, np.eye(2), atol=1e-6), np.fabs(rotation @ rotation.T - np.eye(2))
            assert np.allclose(rotation.T @ rotation, np.eye(2), atol=1e-6), np.fabs(rotation.T @ rotation - np.eye(2))

            cos = (rotation[0, 0] + rotation[1, 1]) / 2.
            sin = (rotation[1, 0] - rotation[0, 1]) / 2.
    
            rotation = np.arctan2(sin, cos)
            
        elif p_from == 'radians':
            assert isinstance(rotation, float)
        else:
            raise ValueError('Parametrization {} not recognized'.format(p_from))

        rotation = rotation % (2*np.pi)
        
        # convert from radians
        if p_to == 'radians':
            pass
        elif p_to == 'MAT':
            cos = np.cos(rotation)
            sin = np.sin(rotation)
            rotation = np.array(([
                [cos, -sin],
                [sin, cos],
            ]))
        else:
            raise ValueError('Parametrization {} not recognized'.format(p_to))
        
        return flip, rotation

    ###########################################################################
    
    def sample(self) -> GroupElement:
        return self.element((np.random.randint(0, 2), np.random.random() * 2 * np.pi))

    def grid(self, N: int, type: str = 'regular', seed: int = None) -> List[GroupElement]:
        r"""
            .. todo::
                Add documentation

        """

        if type == 'regular':
            grid = [
                (0, i * 2*np.pi / N) for i in range(N)
            ] + [
                (1, i * 2 * np.pi / N) for i in range(N)
            ]
        elif type == 'rand':
            if isinstance(seed, int):
                rng = np.random.RandomState(seed)
            elif seed is None:
                rng = np.random
            else:
                assert isinstance(seed, np.random.RandomState)
                rng = seed
            grid = [(rng.randint(0, 2), rng.random()*2*np.pi) for i in range(N)]
        else:
            raise ValueError(f'Grid type {type} not recognized')

        return [
            self.element(g, param='radians')
            for g in grid
        ]

    def grid_so2(self, N: int, type: str = 'regular', seed: int = None) -> List[GroupElement]:
        r"""
            .. todo::
                Add documentation

        """

        if type == 'regular':
            so2_grid = [i * 2*np.pi / N for i in range(N)]
        elif type == 'rand':
            if isinstance(seed, int):
                rng = np.random.RandomState(seed)
            elif seed is None:
                rng = np.random
            else:
                assert isinstance(seed, np.random.RandomState)
                rng = seed
            so2_grid = [rng.random()*2*np.pi for i in range(N)]
        else:
            raise ValueError(f'Grid type {type} not recognized')

        return [self.element((0, g), param='radians') for g in so2_grid]

    def testing_elements(self, n=4*13) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        """
        return iter(
            [self.element((0, i * 2. * np.pi / n)) for i in range(n)]
          + [self.element((1, i * 2. * np.pi / n)) for i in range(n)]
        )
    
    def __eq__(self, other):
        if not isinstance(other, O2):
            return False
        else:
            return self.name == other.name # and self._maximum_frequency == other._maximum_frequency

    def _subgroup(self, id: Tuple[float, int]) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""
        Restrict the current group :math:`O(2)` to the subgroup identified by the input ``id``, where ``id`` is a
        tuple :math:`(\theta, M)`.
        
        Args:
            id (tuple): the identification of the subgroup

        Returns:
            a tuple containing
            
                - the subgroup
                
                - a function which maps an element of the subgroup to its inclusion in the original group and
                
                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)
                
        """
        
        assert isinstance(id, tuple) and len(id) == 2, id
        assert id[0] is None or isinstance(id[0], float), id[0]
        assert isinstance(id[1], int), id[1]

        order = id[1]
        axis = id[0]
        # assert (id[0] is None and (id[1] > 0 or id[1] == -1)) or (id[0] is not None and id[1] > 0)
        # assert axis is None or 0 <= axis < 2*np.pi/order or order < 0, (axis, order)
        assert axis is None or 0 <= axis <= 2*np.pi, (axis, order)
        assert order == -1 or order > 0, (axis, order)
        
        if id[0] is not None and id[1] == -1:
            sg = escnn.group.o2_group(self._maximum_frequency)
            parent_mapping = build_adjoint_map(self, self.element((0, axis / 2)))
            child_mapping = build_adjoint_map(self, self.element((0, -axis / 2)))

        elif id[0] is None and id[1] == -1:
            sg = escnn.group.so2_group(self._maximum_frequency)
            # parent_mapping = lambda e: self.element((0, e._element))
            # child_mapping = lambda e, sg=sg: None if e._element[0] != 0 else sg.element(e._element[1])
            parent_mapping = so2_to_o2(self)
            child_mapping = o2_to_so2(sg)

        elif id[0] is not None and id[1] == 1:
            # take the elements of the group generated by "2pi/k f"
            sg = escnn.group.cyclic_group(2)
            # parent_mapping = lambda e, axis=axis: self.element((e._element, axis * e._element))
            # child_mapping = lambda e, axis=axis, sg=sg: None if not utils.cycle_isclose(e._element[1], axis * e._element[0], 2 * np.pi) else sg.element(e._element[0])
            parent_mapping = flip_to_o2(axis, self)
            child_mapping = o2_to_flip(axis, sg)

        elif id[0] is None:
            # take the elements of the group generated by "2pi/order"
            sg = escnn.group.cyclic_group(order)
            # parent_mapping = lambda e, order=order: self.element((0, e._element * 2. * np.pi / order))
            # child_mapping = lambda e, order=order, sg=sg: None if (e._element[0] != 0 or not utils.cycle_isclose(e._element[1], 0., 2. * np.pi / order)) else \
            #     sg.element(int(round(e._element[1] * order / (2. * np.pi))))
            parent_mapping = so2_to_o2(self)
            child_mapping = o2_to_so2(sg)

        elif id[0] is not None and id[1] > 1:
            # take the elements of the group generated by "2pi/order" and "2pi/k f"
            sg = escnn.group.dihedral_group(order)

            # parent_mapping = lambda e, order=order, axis=axis: self.element((e._element[0], e._element[1] * 2. * np.pi / order + e._element[0] * axis))
            # child_mapping = lambda e, order=order, axis=axis, sg=sg: None if not utils.cycle_isclose(e._element[1] - e._element[0] * axis, 0., 2. * np.pi / order) else \
            #     sg.element((e._element[0], int(round((e._element[1] - e._element[0] * axis) * order / (2. * np.pi)))))
            parent_mapping = dn_to_o2(axis, self)
            child_mapping = o2_to_dn(axis, sg)
        else:
            raise ValueError(f"id '{id}' not recognized")

        return sg, parent_mapping, child_mapping

    def _combine_subgroups(self, sg_id1, sg_id2):
    
        sg_id1 = self._process_subgroup_id(sg_id1)
        sg1, inclusion, restriction = self.subgroup(sg_id1)
        sg_id2 = sg1._process_subgroup_id(sg_id2)
    
        if sg_id1[0] is None:
            return (None, sg_id2)
        elif sg_id1[1] == 1:
            return sg_id1[0] if sg_id2 == 2 else None, 1
        elif sg_id2[0] is None:
            return sg_id2
        else:
            flip = sg_id1[0] + inclusion(sg1.element((0, sg_id2[0]))).to('radians')[1]
            return (flip,) + sg_id2[1:]

    def _restrict_irrep(self, irrep: Tuple, id: Tuple[int, int]) -> Tuple[np.matrix, List[Tuple]]:
        r"""
        Restrict the input irrep of current group to the subgroup identified by "id".
        More precisely, "id" is a tuple :math:`(k, m)`, where :math:`m` is a positive integer indicating the number of
        rotations in the subgroup while :math:`k` is either None (no flips in the subgroup) or an angle in
        :math:`[0, \frac{2\pi}{m})` (indicating the axis of flip in the subgroup).
        Valid combinations are:
        - (None, -1): restrict to the subgroup :math:`SO(2)` containing only the rotations
        - (None, m): restrict to the cyclic subgroup with order "m" :math:`C_m` generated by :math:`\langle r_{2\pi/m} \rangle`.
        - (0, m): restrict to the dihedral subgroup with order "2m" :math:`D_{2m}` generated by :math:`\langle r_{2\pi/m}, f \rangle`
        - (0, 1): restrict to the cyclic subgroup of order 2 :math:`C_2` generated by the flip :math:`\langle f \rangle`
        - (None, 1): restrict to the cyclic subgroup of order 1 :math:`C_1` containing only the identity
        - (\theta, m): restrict to the dihedral subgroup with order "2m" :math:`D_{2m}` generated by :math:`\langle r_{2\pi/m}, r_{\theta} f \rangle`
        
        Args:
            irrep (tuple): the identifier of the irrep to restrict
            id (tuple): the identification of the subgroup

        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep
            
        """

        irr = self.irrep(*irrep)
        
        sg, _, _ = self.subgroup(id)

        irreps = []
        change_of_basis = None

        if id[0] is not None and id[1] == -1:
            
            change_of_basis = np.eye(irr.size)
            irreps.append(irr.id)

            if irr.size == 2:
                f = irr.attributes["frequency"]
                change_of_basis = utils.psi(0.5 * id[0], f) @ change_of_basis

        elif id[0] is None and id[1] == -1:
            f = irr.attributes["frequency"]
            irreps.append((f,))
            change_of_basis = np.eye(irr.size)
            
        elif id[0] is not None and id[1] == 1:
            j = irr.attributes["flip_frequency"]
            k = irr.attributes["frequency"]
            change_of_basis = np.eye(irr.size)
            if irr.size > 1:
                irreps.append((0,))
                change_of_basis = utils.psi(0.5 * id[0], k)
            irreps.append((j,))
        elif id[0] is None and id[1] > 0:
        
            order = id[1]

            f = irr.attributes["frequency"] % order
            if f > order/2:
                f = order - f
                change_of_basis = utils.chi(1)
            else:
                change_of_basis = np.eye(irr.size)

            r = (f,)
            if sg.irrep(*r).size < irr.size:
                irreps.append(r)
            irreps.append(r)
    
        elif id[0] is not None and id[1] > 1:
        
            order = id[1]
            j = irr.attributes["flip_frequency"]
            f = irr.attributes["frequency"]
            k = f % order
            
            if k > order/2:
                k = order - k
                change_of_basis = np.array([[1, 0], [0, -1]])
            else:
                change_of_basis = np.eye(irr.size)
            
            r = (j,k)
            if sg.irrep(*r).size < irr.size:
                irreps.append((0, k))
            irreps.append(r)
                
            if irr.size == 2:
                change_of_basis = utils.psi(0.5 * id[0], f) @ change_of_basis

        else:
            raise ValueError(f"id '{id}' not recognized")
        
        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps for this group

        """
        
        # Build all the Irreducible Representations
    
        j, k = 0, 0
    
        # add Trivial representation
        self.irrep(j, k)
    
        j = 1
        for k in range(self._maximum_frequency + 1):
            self.irrep(j, k)

        # Build all Representations
        
        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name : irr for irr in self.irreps()})

    def bl_regular_representation(self, L: int) -> Representation:
        r"""
        Band-Limited regular representation up to frequency ``L`` (included).

        Args:
            L(int): max frequency

        """
        name = f'regular_{L}'
    
        if name not in self._representations:
            irreps = []
        
            for l in range(L + 1):
                if l == 0:
                    irreps += [self.irrep(0, l)]

                irreps += [self.irrep(1, l)] * self.irrep(1, l).size
        
            self._representations[name] = directsum(irreps, name=name)
    
        return self._representations[name]

    def bl_quotient_representation(self,
                                   L: int,
                                   subgroup_id,
                                   name: str = None,
                                   ) -> escnn.group.Representation:
        r"""
        Band-Limited quotient representation up to frequency ``L`` (included).

        The quotient representation corresponds to the action of the current group :math:`G` on functions over the
        homogeneous space :math:`G/H`, where :math:`H` is the subgroup of :math:`G` identified by ``subgroup_id``.

        Args:
            L(int): max frequency
            subgroup_id: id identifying the subgroup H.
            name (str, optional)

        """

        if name is None:
            name = f"quotient[{subgroup_id}]_{L}"

        if name not in self.representations:
            subgroup, _, _ = self.subgroup(subgroup_id)

            homspace = self.homspace(subgroup_id)

            irreps = []

            irreps_ids = [(0,0)] + [(1, l) for l in range(L+1)]

            for id in irreps_ids:
                irr = self.irrep(*id)
                multiplicity = homspace.dimension_basis(irr.id, homspace.H.trivial_representation.id)[1]
                irreps += [irr] * multiplicity

            self.representations[name] = directsum(irreps, name=name)

        return self.representations[name]

    def bl_irreps(self, L: int) -> List[Tuple]:
        r"""
        Returns a list containing the id of all irreps of (rotational) frequency smaller or equal to ``L``.
        This method is useful to easily specify the irreps to be used to instantiate certain objects, e.g. the
        Fourier based non-linearity :class:`~escnn.nn.FourierPointwise`.
        """
        assert 0 <= L, L
        irreps = [(0, 0)]
        for l in range(L + 1):
            irreps += [(1, l)]
        return irreps

    @property
    def trivial_representation(self) -> Representation:
        return self.representations['irrep_0,0']

    def standard_representation(self) -> Representation:
        r"""
        Standard representation of :math:`\O2` as 2x2 rotation matrices

        This is equivalent to ``self.irrep(1, 1)``.

        """
        return self.irrep(1, 1)

    def irrep(self, j: int, k: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep with reflection and rotation frequencies :math:`j` (reflection) and :math:`k` (rotation) of the
        current group.
        Notice: the frequencies has to be non-negative integers: :math:`j \in \{0, 1\}` and :math:`k \in \mathbb{N}`
        
        Valid parameters are :math:`(0, 0)` and :math:`(1, 0)`, :math:`(1, 1)`, :math:`(1, 2)`, :math:`(1, 3)`, ...
        
        Args:
            j (int): the frequency of the reflection in the irrep
            k (int): the frequency of the rotations in the irrep

        Returns:
            the corresponding irrep

        """
    
        assert j in [0, 1]
        assert k >= 0
    
        name = f"irrep_{j},{k}"
        id = (j, k)

        if id not in self._irreps:
            irrep = _build_irrep_o2(j, k)
            character = _build_char_o2(j, k)
            
            if j == 0:
                if k == 0:
                    # Trivial representation
                    supported_nonlinearities = ['pointwise', 'norm', 'gated', 'gate']
                    self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                                  supported_nonlinearities=supported_nonlinearities,
                                                                  character=character,
                                                                  # trivial=True,
                                                                  frequency=k,
                                                                  flip_frequency=j
                                                                  )
                else:
                    raise ValueError(f"Error! Flip frequency {j} and rotational frequency {k} don't correspond to any irrep of the group {self.name}!")
                
            elif k == 0:

                # add Trivial on SO(2) subgroup Representation
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=k,
                                                              flip_frequency=j
                                                              )
            else:
                # 2 dimensional Irreducible Representations
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 2, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=k,
                                                              flip_frequency=j
                                                              )

        return self._irreps[id]

    def _induced_from_irrep(self,
                            subgroup_id: Tuple[float, int],
                            repr: IrreducibleRepresentation,
                            representatives: List[GroupElement] = None,
                            ) -> Tuple[List[IrreducibleRepresentation], np.ndarray, np.ndarray]:
        
        if representatives is None:
            if subgroup_id == (None, -1):
                # SO(2) has finite index in O(2)
                # so we can build the set of representatives explicitly
                representatives = [self.identity, self.reflection]
            else:
                raise ValueError(f"Induction from discrete subgroups of O(2) leads to infinite dimensional induced "
                                 f"representations. Hence, induction from the subgroup identified "
                                 f"by {subgroup_id} is not allowed.")
        
        return super(O2, self)._induced_from_irrep(subgroup_id, repr, representatives)

    _cached_group_instance = None

    @classmethod
    def _generator(cls, maximum_frequency: int = 3) -> 'O2':
        if cls._cached_group_instance is None:
            cls._cached_group_instance = O2(maximum_frequency)
        elif cls._cached_group_instance._maximum_frequency < maximum_frequency:
            cls._cached_group_instance._maximum_frequency = maximum_frequency
            cls._cached_group_instance._build_representations()
    
        return cls._cached_group_instance


def _build_irrep_o2(j: int, k: int):
    
    def irrep(element: GroupElement, j=j, k=k) -> np.ndarray:
        assert j in [0, 1]
        assert k >= 0
        
        if j == 0:
            if k == 0:
                # Trivial representation
                return np.eye(1)
            else:
                raise ValueError(
                    f"Error! Flip frequency {j} and rotational frequency {k} don't correspond to any irrep of the group {element.group.name}!")
        elif k == 0:
            # Trivial on SO(2) subgroup Representation
            return np.array([[-1 if element.to('radians')[0] else 1]])
        else:
            e = element.to('radians')
            # 2 dimensional Irreducible Representations
            return utils.psichi(e[1], e[0], k=k)
        
    return irrep


def _build_char_o2(j: int, k: int):
    
    def character(element: GroupElement, j=j, k=k) -> float:
        assert j in [0, 1]
        assert k >= 0
        
        if j == 0:
            if k == 0:
                # Trivial representation
                return 1.
            else:
                raise ValueError(
                    f"Error! Flip frequency {j} and rotational frequency {k} don't correspond to any irrep of the group {element.group.name}!")
        elif k == 0:
            # add Trivial on SO(2) subgroup Representation
            return -1 if element.to('radians')[0] else 1
        else:
            e = element.to('radians')
            # 2 dimensional Irreducible Representations
            return 0 if e[0] else (2 * np.cos(k * e[1]))
        
    return character


# SO2 (and Cyclic) ###############################

def o2_to_so2(so2: Union[escnn.group.SO2, escnn.group.CyclicGroup]):

    def _map(e: GroupElement, so2=so2):
        assert isinstance(e.group, O2)

        flip, rotation = e.to('radians')

        if flip == 0:
            try:
                return so2.element(rotation, 'radians')
            except ValueError:
                return None
        else:
            return None

    return _map


def so2_to_o2(o2: O2):

    def _map(e: GroupElement, o2=o2):
        assert isinstance(e.group, escnn.group.SO2) or isinstance(e.group, escnn.group.CyclicGroup)
        return o2.element(
            (0, e.to('radians')), 'radians'
        )

    return _map


# Flip wrt an axis ######################################


def o2_to_flip(axis: float, flip: escnn.group.CyclicGroup):
    assert isinstance(flip, escnn.group.CyclicGroup) and flip.order() == 2

    def _map(e: GroupElement, flip=flip, axis=axis):
        assert isinstance(e.group, O2)

        f, rot = e.to('radians')

        if f == 0 and utils.cycle_isclose(rot, 0, 2*np.pi):
            return flip.identity
        elif f == 1 and utils.cycle_isclose(rot, axis, 2*np.pi):
            return flip.element(1)
        else:
            return None

    return _map


def flip_to_o2(axis: float, o2: O2):

    def _map(e: GroupElement, axis=axis, o2=o2):
        assert isinstance(e.group, escnn.group.CyclicGroup) and e.group.order() == 2

        f = e.to('int')

        if f == 0:
            return o2.identity
        else:
            return o2.element((1, axis))

    return _map


# Dihedral Group ######################################


def o2_to_dn(axis: float, dn: escnn.group.DihedralGroup):
    assert isinstance(dn, escnn.group.DihedralGroup)

    def _map(e: GroupElement, dn=dn, axis=axis):
        assert isinstance(e.group, O2)

        f, rot = e.to('radians')

        if utils.cycle_isclose(rot - f * axis, 0., 2. * np.pi / dn.rotation_order):
            return dn.element((f, int(round((rot - f * axis) * dn.rotation_order / (2. * np.pi)))))
        else:
            return None

    return _map


def dn_to_o2(axis: float, o2: O2):

    def _map(e: GroupElement, axis=axis, o2=o2):
        assert isinstance(e.group, escnn.group.DihedralGroup)

        f, rot = e.to('int')

        return o2.element((f, rot * 2. * np.pi / e.group.rotation_order + f * axis))

    return _map







from escnn.group import direct_product

from .cyclicgroup import CyclicGroup
from .dihedralgroup import DihedralGroup
from .so2group import SO2
from .o2group import O2
from .so3group import SO3
from .o3group import O3
from .ico import Icosahedral
from .octa import Octahedral


__all__ = [
    "trivial_group",
    "cyclic_group",
    "dihedral_group",
    "so2_group",
    "o2_group",
    "so3_group",
    "o3_group",
    "ico_group",
    "octa_group",
    "full_ico_group",
    "full_octa_group",
    "full_cylinder_group",
    "cylinder_group",
    "full_cylinder_discrete_group",
    "cylinder_discrete_group",
]


def trivial_group():
    r"""
    
    Builds the trivial group :math:`C_1` which contains only the identity element :math:`e`.
    
    You should use this factory function to build an instance of the trivial group.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built,
    this unique instance is updated with the new representations and, therefore, all its references will see the new
    representations.
    
    Returns:
        the trivial group

    """
    return CyclicGroup._generator(1)


def cyclic_group(N: int):
    r"""

    Builds a cyclic group :math:`C_N`of order ``N``, i.e. the group of ``N`` discrete planar rotations.
    
    You should use this factory function to build an instance of :class:`escnn.group.CyclicGroup`.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built
    (eg. through the method :meth:`~escnn.group.Group.quotient_representation`), this unique instance is updated with
    the new representations and, therefore, all its references will see the new representations.

    Args:
        N (int): number of discrete rotations in the group

    Returns:
        the cyclic group of order ``N``

    """
    return CyclicGroup._generator(N)


def dihedral_group(N: int):
    r"""

    Builds a dihedral group :math:`D_{2N}`of order ``2N``, i.e. the group of ``N`` discrete planar rotations
    and reflections.
    
    You should use this factory function to build an instance of :class:`escnn.group.DihedralGroup`.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built
    (eg. through the method :meth:`~escnn.group.Group.quotient_representation`), this unique instance is updated with
    the new representations and, therefore, all its references will see the new representations.

    Args:
        N (int): number of discrete rotations in the group
        
    Returns:
        the dihedral group of order ``2N``

    """
    return DihedralGroup._generator(N)


def so2_group(maximum_frequency: int = 3):
    r"""

    Builds the group :math:`SO(2)`, i.e. the group of continuous planar rotations.
    Since the group has infinitely many irreducible representations, it is not possible to build all of them.
    Each irrep is associated to one unique frequency and the parameter ``maximum_frequency`` specifies
    the maximum frequency of the irreps to build.
    New irreps (associated to higher frequencies) can be manually created by calling the method
    :meth:`escnn.group.SO2.irrep` (see the method's documentation).
    
    You should use this factory function to build an instance of :class:`escnn.group.SO2`.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built
    (eg. through the method :meth:`escnn.group.SO2.irrep`), this unique instance is updated with the new representations
    and, therefore, all its references will see the new representations.
    
    Args:
        maximum_frequency (int): maximum frequency of the irreps

    Returns:
        the group :math:`SO(2)`

    """
    return SO2._generator(maximum_frequency)


def o2_group(maximum_frequency: int = 3):
    r"""

    Builds the group :math:`O(2)`, i.e. the group of continuous planar rotations and reflections.
    Since the group has infinitely many irreducible representations, it is not possible to build all of them.
    Each irrep is associated to one unique frequency and the parameter ``maximum_frequency`` specifies
    the maximum frequency of the irreps to build.
    New irreps (associated to higher frequencies) can be manually created by calling the method
    :meth:`escnn.group.O2.irrep` (see the method's documentation).
    
    You should use this factory function to build an instance of :class:`escnn.group.O2`.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built
    (eg. through the method :meth:`escnn.group.O2.irrep`), this unique instance is updated with the new representations
    and, therefore, all its references will see the new representations.
    
    Args:
        maximum_frequency (int): maximum frequency of the irreps

    Returns:
        the group :math:`O(2)`

    """
    return O2._generator(maximum_frequency)


def so3_group(maximum_frequency: int = 2):
    r"""

    Builds the group :math:`SO(3)`, i.e. the group of continuous 3D rotations.
    
    Args:
        maximum_frequency (int): maximum frequency of the irreps

    Returns:
        the group :math:`SO(3)`

    """
    return SO3._generator(maximum_frequency)


def o3_group(maximum_frequency: int = 2):
    r"""

    Builds the group :math:`O(3)`, i.e. the group of continuous 3D rotations and reflections.
    
    Args:
        maximum_frequency (int): maximum frequency of the irreps

    Returns:
        the group :math:`O(3)`

    """
    return O3._generator(maximum_frequency)


def ico_group():
    r"""

    Builds the group :math:`I`, i.e. the group of symmetries of the icosahedron

    Returns:
        the group :math:`I`

    """
    return Icosahedral._generator()


def full_ico_group():
    r"""

    Builds the group :math:`I_h = C_2 \times I`, i.e. the group of all symmetries of the icosahedron

    Returns:
        the group :math:`I`

    """
    ico = ico_group()
    c2 = cyclic_group(2)
    return direct_product(c2, ico, name='FullIcosahedral')


def octa_group():
    r"""

    Builds the group :math:`O`, i.e. the group of symmetries of the cube or octahedron

    Returns:
        the group :math:`O`

    """
    return Octahedral._generator()


def full_octa_group():
    r"""

    Builds the group :math:`O_h = C_2 \times O`, i.e. the group of all symmetries of the cube or octahedron

    Returns:
        the group :math:`O_h`

    """
    octa = octa_group()
    c2 = cyclic_group(2)
    return direct_product(c2, octa, name='FullOctahedral')


def full_cylinder_group(maximum_frequency: int = 3):
    r"""

    Builds the group :math:`C_2 \times O(2)`, i.e. the group of symmetries of a cylinder.
    It contains continuous planar rotations and planar reflections (the :math:`O(2)` dihedral symmetry subgroup)
    and 3D inversions (the :math:`C_2` subgroups).
    The group also includes the subgroup of mirroring with respect to a plane passing throgh the axis of the cylinder.
    
    Since the group has infinitely many irreducible representations, it is not possible to build all of them.
    Each irrep is associated to one unique frequency and the parameter ``maximum_frequency`` specifies
    the maximum frequency of the irreps to build.
    New irreps (associated to higher frequencies) can be manually created by calling the method
    :meth:`~escnn.group.DirectProductGroup.irrep` (see the method's documentation).

    You should use this factory function to build an instance of of this group.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built
    (eg. through the method :meth:`~escnn.group.O2.irrep` of an instance of :class:`~escnn.group.O2`), this unique
    instance is updated with the new representations and, therefore, all its references will see the new representations.

    Args:
        maximum_frequency (int): maximum frequency of the :math:`O(2)` irreps

    Returns:
        the group :math:`C_2 \times O(2)`

    """
    o2 = o2_group(maximum_frequency)
    c2 = cyclic_group(2)
    
    return direct_product(c2, o2, name='FullCylindrical')


def cylinder_group(maximum_frequency: int = 3):
    r"""

    Builds the group :math:`C_2 \times SO(2)`, which contains continuous planar rotations (the :math:`SO(2)`
    symmetry subgroup) and 3D inversions (the :math:`C_2` subgroups).

    Since the group has infinitely many irreducible representations, it is not possible to build all of them.
    Each irrep is associated to one unique frequency and the parameter ``maximum_frequency`` specifies
    the maximum frequency of the irreps to build.
    New irreps (associated to higher frequencies) can be manually created by calling the method
    :meth:`~escnn.group.DirectProductGroup.irrep` (see the method's documentation).

    You should use this factory function to build an instance of of this group.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.
    In case of multiple calls of this function with different parameters or in case new representations are built
    (eg. through the method :meth:`~escnn.group.SO2.irrep` of an instance of :class:`~escnn.group.SO2`), this unique
    instance is updated with the new representations and, therefore, all its references will see the new representations.

    Args:
        maximum_frequency (int): maximum frequency of the :math:`SO(2)` irreps

    Returns:
        the group :math:`C_2 \times O(2)`

    """
    so2 = so2_group(maximum_frequency)
    c2 = cyclic_group(2)
    
    return direct_product(c2, so2, name='Cylindrical')


def full_cylinder_discrete_group(n: int):
    r"""

    Builds the group :math:`C_2 \times D_n`, i.e. a group of discrete symmetries of a cylinder.
    It contains `n` discrete planar rotations and planar reflections (the :math:`D_n` dihedral symmetry subgroup)
    and 3D inversions (the :math:`C_2` subgroups).
    The group also includes the subgroup of mirroring with respect to `n` planes passing through the axis of the
    cylinder.

    You should use this factory function to build an instance of of this group.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.

    Args:
        n (int): number of discrete rotations

    Returns:
        the group :math:`C_2 \times D_n`

    """
    dn = dihedral_group(n)
    c2 = cyclic_group(2)
    
    return direct_product(c2, dn, name='FullCylindricalDiscrete')


def cylinder_discrete_group(n: int):
    r"""

    Builds the group :math:`C_2 \times C_n`, which contains `n` discrete planar rotations (the :math:`C_n`
    symmetry subgroup) and 3D inversions (the :math:`C_2` subgroups).

    You should use this factory function to build an instance of of this group.
    Only one instance is built and, in case of multiple calls to this function, the same instance is returned.

    Args:
        n (int): number of discrete rotations

    Returns:
        the group :math:`C_2 \times C_n`

    """
    cn = cyclic_group(n)
    c2 = cyclic_group(2)
    
    return direct_product(c2, cn, name='CylindricalDiscrete')





from collections import defaultdict

from torch.nn import Parameter

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
import torch.nn.functional as F

from typing import List, Tuple, Any, Union

import math


__all__ = ["NormMaxPool"]


class NormMaxPool(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 dilation: Union[int, Tuple[int, int]] = 1,
                 ceil_mode: bool = False
                 ):
        r"""
        
        Max-pooling based on the fields' norms. In a given window of shape :attr:`kernel_size`, for each
        group of channels belonging to the same field, the field with the highest norm (as the length of the vector)
        is preserved.
        
        Except :attr:`in_type`, the other parameters correspond to the ones of :class:`torch.nn.MaxPool2d`.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a max over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            dilation: a parameter that controls the stride of elements in the window
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape
            
        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 2

        super(NormMaxPool, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        if isinstance(kernel_size, int):
            self.kernel_size = (kernel_size, kernel_size)
        else:
            self.kernel_size = kernel_size
        
        if isinstance(stride, int):
            self.stride = (stride, stride)
        elif stride is None:
            self.stride = self.kernel_size
        else:
            self.stride = stride
        
        if isinstance(padding, int):
            self.padding = (padding, padding)
        else:
            self.padding = padding
        
        if isinstance(dilation, int):
            self.dilation = (dilation, dilation)
        else:
            self.dilation = dilation
            
        self.ceil_mode = ceil_mode
        
        self._nfields = None
        
        # group fields by their size and
        #   - check if fields of the same size are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each size
        self._nfields = defaultdict(int)
        
        # indices of the channales corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        position = 0
        last_size = None
        for i, r in enumerate(self.in_type.representations):
            
            if r.size != last_size:
                if not r.size in self._contiguous:
                    self._contiguous[r.size] = True
                else:
                    self._contiguous[r.size] = False
            last_size = r.size
            
            _indices[r.size] += list(range(position, position + r.size))
            self._nfields[r.size] += 1
            position += r.size
        
        for s, contiguous in self._contiguous.items():
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[s] = torch.LongTensor([min(_indices[s]), max(_indices[s])+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[s] = torch.LongTensor(_indices[s])
                
            # register the indices tensors as parameters of this module
            self.register_buffer('indices_{}'.format(s), _indices[s])
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Run the norm-based max-pooling on the input tensor
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        b, c, hi, wi = input.tensor.shape
        
        # compute the output shape (see 'torch.nn.MaxPool2D')
        b, c, ho, wo = self.evaluate_output_shape(input.tensor.shape)
        
        # compute the squares of the values of each channel
        # n = torch.mul(input.data, input.data)
        n = input.tensor ** 2
        
        # pre-allocate the output tensor
        output = torch.empty(b, c, ho, wo, device=input.tensor.device)
        
        # reshape the input to merge the spatial dimensions
        input = input.tensor.reshape(b, c, -1)
        
        # iterate through all field sizes
        for s, contiguous in self._contiguous.items():
            indices = getattr(self, f"indices_{s}")
            
            if contiguous:
                # if the fields were contiguous, we can use slicing
                
                # compute the norms
                norms = n[:, indices[0]:indices[1], :, :]\
                        .view(b, -1, s, hi, wi)\
                        .sum(dim=2)\
                        .sqrt()
                
                # run max-pooling on the norms-tensor
                _, indx = F.max_pool2d(norms,
                                       self.kernel_size,
                                       self.stride,
                                       self.padding,
                                       self.dilation,
                                       self.ceil_mode,
                                       return_indices=True)
                
                # in order to use the pooling indices computed for the norms to retrieve the fields, they need to be
                # expanded in the inner field dimension
                indx = indx.view(b, -1, 1, ho * wo).expand(-1, -1, s, -1)
                
                # retrieve the fields from the input tensor using the pooling indeces
                output[:, indices[0]:indices[1], :, :] = input[:, indices[0]:indices[1], :]\
                                                              .view(b, -1, s, hi*wi)\
                                                              .gather(3, indx)\
                                                              .view(b, -1, ho, wo)
                
            else:
                # otherwise we have to use indexing
                
                # compute the norms
                norms = n[:, indices, :, :] \
                    .view(b, -1, s, hi, wi) \
                    .sum(dim=2) \
                    .sqrt()
                
                # run max-pooling on the norms-tensor
                _, indx = F.max_pool2d(norms,
                                       self.kernel_size,
                                       self.stride,
                                       self.padding,
                                       self.dilation,
                                       self.ceil_mode,
                                       return_indices=True)
                
                # in order to use the pooling indices computed for the norms to retrieve the fields, they need to be
                # expanded in the inner field dimension
                indx = indx.view(b, -1, 1, ho * wo).expand(-1, -1, s, -1)

                # retrieve the fields from the input tensor using the pooling indeces
                output[:, indices, :, :] = input[:, indices, :] \
                    .view(b, -1, s, hi * wi) \
                    .gather(3, indx) \
                    .view(b, -1, ho, wo)
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type.size

        b, c, hi, wi = input_shape

        # compute the output shape (see 'torch.nn.MaxPool2D')
        ho = math.floor(
            (hi + 2 * self.padding[0] - self.dilation[0] * (self.kernel_size[0] - 1) - 1) / self.stride[0] + 1)
        wo = math.floor(
            (wi + 2 * self.padding[1] - self.dilation[1] * (self.kernel_size[1] - 1) - 1) / self.stride[1] + 1)

        return b, self.out_type.size, ho, wo
        
    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        # this kind of pooling is not really equivariant so we can not test equivariance
        pass
